/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Louis-Marie Traonouez <louis-marie.traonouez@inria.fr>
*/

grammar Preprocessor;

@parser::header {
    package preprocessor;
    
    import java.math.BigInteger;
    import java.math.BigDecimal;
    import java.util.Iterator;
    import java.util.LinkedList;
}

@lexer::header {
    package preprocessor; 
}


@members {
    Programme prog;
}

prog_grammar[Programme p]: { prog=p; }  decl* st_list[null] EOF  { prog.addStatements($st_list.l); } ;

decl:	const_decl
    |	public_decl 
    |	private_decl
    ;

const_decl:	'const' IDENTIFIER ':=' 
		( val=assign_expr[false,null]   
		  {
		      try {
			  BigDecimal int_val = new BigDecimal($val.str);
			  if (!prog.addConstant($IDENTIFIER.text,int_val)) {
			      System.err.println("WARNING: on line " + $IDENTIFIER.getLine() + ", constant " + $IDENTIFIER.text + " is already declared.");
			  }
		      }
		      catch(NumberFormatException e) {
			  System.err.println("ERROR: on line " + $IDENTIFIER.getLine() + ", " + $val.str + " cannot be converted to integer.");
			  System.exit(1);
		      }
		  }
		  
		  | fval=float_number
		  {
		      if (!prog.addConstant($IDENTIFIER.text,$fval.f)) {
			  System.err.println("WARNING: on line " + $IDENTIFIER.getLine() + ", constant " + $IDENTIFIER.text + " is already declared.");
		      }
		  }
		)
                ';'
	  ;
	  
public_decl:	public_type var_type[true] IDENTIFIER init_val ';'
		{   $var_type.v.setPublicity($public_type.b);
		    if (!prog.addPublicVariable($IDENTIFIER.text,((PublicVariable)$var_type.v))) {
			System.err.println("WARNING: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is already declared.");
		    }
		    else{
    			$var_type.v.setName($IDENTIFIER.text);
			prog.addDeclaration($IDENTIFIER.text);
			if (!$init_val.l.isEmpty()) {
			    int r = ((PublicVariable)$var_type.v).updateInitialValues($init_val.l);
			    if (r>0) {
				System.err.println(" on line " + $IDENTIFIER.getLine() + ".");
				if (r==2) {
				    System.exit(1);
				}
			    }
			}
		    }
		}
	   ;

public_type returns [boolean b]:
	   'public' 		{ $b = true; }
	   | 'observable' 	{ $b = false; }
	   ;
	   
private_decl:	private_type var_type[false] IDENTIFIER init_intervals ';'
		{   $var_type.v.setPublicity($private_type.b);
		    if (!prog.addPrivateVariable($IDENTIFIER.text,((PrivateVariable)$var_type.v))) {
			System.err.println("WARNING: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is already declared.");
		    }
		    else {
			$var_type.v.setName($IDENTIFIER.text);
			prog.addDeclaration($IDENTIFIER.text);
			if (!$init_intervals.l.isEmpty()) {
			    int r = ((PrivateVariable)$var_type.v).updateIntRange($init_intervals.l);
			    if (r>0) {
				System.err.println(" on line " + $IDENTIFIER.getLine() + ".");
				if (r==2) {
				    System.exit(1);
				}
			    }
			}
		    }
		}
	    ;
	    
private_type returns [boolean b]:
	    'private'		{ $b = true; }
	    | 	'secret' 	{ $b = false; }
	    ;

var_type[boolean pub] returns [Variable v] :
	INT
	{
	    String sizeString = $INT.text.substring(3);
	    int sizeInt = 0;
	    try {
		sizeInt = Integer.parseInt(sizeString);
	    }
	    catch (NumberFormatException e) {
		BigDecimal val = prog.getConstantValue(sizeString);
		if (val != null)
		    sizeInt = val.intValue();
		else
		    throw new RuntimeException("Unknown constant: " + sizeString);
	    }    
	    if (pub) {
		$v = new PublicVariable(sizeInt);
	    }
	    else { 
		$v = new PrivateVariable(sizeInt);
	    }
	}
	| a='array' '[' e1=assign_expr[false,null] ('..' e2=assign_expr[false,null])? ']' 'of' w=var_type[$pub]
	{
	    String smin = "0";
	    String smax = $e1.str;

	    if (e2 != null) {
		smin = smax;
		smax = $e2.str;
	    }
	    $v = $w.v;
	    try {
		int min = Integer.parseInt(smin);
		int max = Integer.parseInt(smax);
		if (e2 == null) {
		    max -= 1;
		}
		$v.addDimension(new ArrayDimension(min,max));
	    }
	    catch(NumberFormatException e) {
		System.err.println("ERROR: on line " + a.getLine() + ", array size " + smin + " or " + smax + " cannot be converted to integer.");
		System.exit(1);
	    }
	}
	;
   
init_val returns [LinkedList<BigInteger> l]
	@init{ l = new LinkedList<BigInteger>(); } :
	// empty
	| a=':='( val=assign_expr[false,null] 			{   try {
									BigInteger int_val = new BigInteger($val.str);
									$l.add(int_val);
								    }
								    catch(NumberFormatException e) {
									System.err.println("ERROR: on line " + a.getLine() + ", " + $val.str + " cannot be converted to integer.");
									System.exit(1);
								    }
								} 
		| '{' number_list '}'				{   $l = $number_list.l; }
		)
	;

number_list returns [LinkedList<BigInteger> l]
	@init{ l = new LinkedList<BigInteger>(); } :
	val1=assign_expr[false,null] (a=',' val2=assign_expr[false,null]
			  {
			      try {
				  BigInteger int_val = new BigInteger($val2.str);
				  $l.add(int_val);
			      }
			      catch(NumberFormatException e) {
				  System.err.println("ERROR: on line " + a.getLine() + ", " + $val2.str + " cannot be converted to integer.");
				  System.exit(1);
			      }
			  }
			 )*
	{
	    try {
		BigInteger int_val = new BigInteger($val1.str);
		$l.addFirst(int_val);
            }
	    catch(NumberFormatException e) {
		System.err.println("ERROR: on line " + a.getLine() + ", " + $val1.str + " cannot be converted to integer.");
		System.exit(1);
	    }
	}
	;

init_intervals returns [LinkedList<BigInteger> l] 
	@init{ l = new LinkedList<BigInteger>(); } :
	// empty
	| a=':=' (interval
		  { 
		      if ($l.isEmpty()) {
			  $l.addAll($interval.l);
		      }
		      else {
			  int comp = $l.getLast().compareTo($interval.l.getFirst().subtract(BigInteger.ONE));
			  if (comp < 0) {
			      $l.add($interval.l.getFirst());
			      $l.add($interval.l.getLast());
			  }
			  else {
			      comp = $l.getLast().compareTo($interval.l.getFirst());
			      if (comp <= 0) {
				  $l.removeLast();
				  $l.add($interval.l.getLast());
			      }
			      else {
				  System.err.println("ERROR: on line " + $a.getLine() + ", overlapping intervals.");
				  System.exit(1);
			      }
			  }	    
		      }
		  }
		 )+
	;
	
interval returns [LinkedList<BigInteger> l]:
	a='[' min=assign_expr[false,null] ',' max=assign_expr[false,null] ']'
	{
	    try {
		BigInteger minVal = new BigInteger($min.str);
		BigInteger maxVal = new BigInteger($max.str);
		$l = new LinkedList<BigInteger>();
		if (minVal.compareTo(maxVal) > 0) {
		    System.err.println("WARNING: on line " + $a.getLine() + ", interval [" + minVal.toString() + "," + maxVal.toString() + "] is empty.");
		}
		else {
		    $l.add(minVal);
    		    $l.add(maxVal);
		}
            }
	    catch(NumberFormatException e) {
		System.err.println("ERROR: on line " + $a.getLine() + ", " + $min.str + " or " + $max.str + " cannot be converted to integer.");
		System.exit(1);
	    }
	}
	;

st_list [Statement parent] returns [LinkedList<Statement> l]
@init { 
  $l = new LinkedList<Statement>(); 
} :
( st[parent] 			{ $l.addAll($st.l); } 
)*
;
	
st [Statement parent] returns [LinkedList<Statement> l]:
    assignement_st[parent]	{ $l = new LinkedList<Statement>(); $l.add($assignement_st.s); }
    | return_st 		{ $l = new LinkedList<Statement>(); $l.add($return_st.s); }
    | random_st[parent]		{ $l = new LinkedList<Statement>(); $l.add($random_st.s); }
    | while_st[parent] 		{ $l = new LinkedList<Statement>(); $l.add($while_st.ifst); }
    | for_st[parent]		{ $l = $for_st.l; }
    | if_st[parent] 		{ $l = new LinkedList<Statement>(); $l.add($if_st.ifst); }
    ;
  
assignement_st [Statement parent] returns [AssignStatement s]
@init {
  $s = new AssignStatement();
  if (parent == null)
    parent = $s;
} :
'assign' var=variable[false, false, parent] ':=' e=assign_expr[false,parent] ';' 
{
    $s.setVariable(new String("(" + $var.str + ")"));
    $s.setExpression(new String("(" + $e.str + ")"));
}
;

return_st returns [ReturnStatement s]:
'return' ';' 			{ $s = new ReturnStatement(); }
;

random_st [Statement parent] returns [RandomStatement s]
@init {
  $s = new RandomStatement();
  if (parent == null)
    parent = $s;
} :
ran='random' var=variable[false, false, parent] ':=' 	{ $s.setVariable(new String("(" + $var.str + ")")); }
( 'randombit' '(' 
  ( 
      float_number 
      {
	  if ($float_number.f.signum() < 0 || $float_number.f.compareTo(BigDecimal.ONE) == 1) {
	    System.err.println( "ERROR: on line " + $ran.getLine() + ", float value " +  $float_number.f.toPlainString() + " must be in [0,1].");
	    System.exit(1);
	  }
	  
	  $s.setProbability($float_number.f);
      }
  
  |   IDENTIFIER
      {
	  BigDecimal val = prog.getConstantValue($IDENTIFIER.text);
	  if (val != null) {
	      if (val.signum() < 0 || val.compareTo(BigDecimal.ONE) == 1) {
		  System.err.println( "ERROR: on line " + $ran.getLine() + ", float value " +  val.toPlainString() + " must be in [0,1].");
		  System.exit(1);
	      }
	      $s.setProbability(val);
	  }
	  else {
	      throw new RuntimeException("Unknown constant: " + $IDENTIFIER.text);
	  }
      }
  )
  ')'
  
| 'random' '(' e=assign_expr[false,parent] ',' f=assign_expr[false,parent] ')'
  {
      $s.setExpression(new String("(" + $e.str + "),(" + $f.str +')'));
  }
)
';'
;

while_st [Statement parent] returns [IfStatement ifst]
@init {
  $ifst = new IfStatement();
  if (parent == null)
    parent = $ifst;
} :
'while' '(' assign_expr[true,parent] ')' 'do' st_list[parent] 'od' 
{
    $ifst.setGuard($assign_expr.str);
    $ifst.setThenGoto(new GotoStatement(2));
    for (Statement st : $st_list.l) {
	$ifst.addThenStatement(st);
    }
    $ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
    $ifst.setExitGoto(new GotoStatement((-1)*$ifst.getNbThenStatements()-3));
}
;
    

for_st [Statement parent] returns [LinkedList<Statement> l]
@init {
    Variable baseTabVar;
} :
'for' '(' var=IDENTIFIER 'in' tab=IDENTIFIER indices[parent] ')'
{
    baseTabVar = prog.getVariable($tab.text);
    if (baseTabVar == null) {
        System.err.println( "ERROR: on line " + $tab.getLine() + ", variable " +  $tab.text + " is not declared.");
        System.exit(1);
    }
    
    ArrayDimension tabVar = baseTabVar.getArrayDimension($indices.l.size());
    if (tabVar == null) {
        System.err.println( "ERROR: on line " + $tab.getLine() + ", variable " +  $tab.text + " has not enough dimensions.");
        System.exit(1);
    }
    
    PublicVariable v = new PublicVariable(baseTabVar.getIntType());
    if (!prog.addPublicVariable($var.text,v)) {
        System.err.println("WARNING: on line " + $var.getLine() + ", variable " +  $var.text + " is already declared.");
    }
    for (int i = 1+$indices.l.size(); i < baseTabVar.getDimensions(); i++) {
        v.addDimension(baseTabVar.getArrayDimension(i));
    }
}
'do' st_list[parent] 'od'
{
    $l = new LinkedList<Statement>();
    for (int i = tabVar.getMinIndex(); i <= tabVar.getMaxIndex(); i++) {
		    String newTabIndex = new String($tab.text);
		    for (Iterator<String> it = $indices.l.iterator(); it.hasNext();) {
		        String s = it.next();
		        newTabIndex += s;
				}
				newTabIndex += "["+i+"]";
				for (Iterator<Statement> it = $st_list.l.iterator(); it.hasNext();) {
				    Statement s = it.next();
				    s = s.replace($var.text,newTabIndex);
				    if (parent == null)
				        parent = s;
				    if (baseTabVar.getPublicity() == 1) {
				        baseTabVar.setLastUsed(parent);
				        parent.addUsedVariable(baseTabVar);
				    }
            if (parent == s)
                parent = null;
				    $l.add(s);
				}
    }
    prog.removeVariable($var.text);
}
;

if_st [Statement parent] returns [IfStatement ifst]
@init {
    ifst = new IfStatement();
    if (parent == null)
        parent = ifst;
} :
if_case[parent,ifst] (else_cases[parent,ifst])? 'fi'
{
    $ifst.setExitGoto(new GotoStatement($ifst.getNbElseStatements()+1));
} 
;

if_case [Statement parent, IfStatement ifst] :
    'if' '(' assign_expr[true,parent] ')' 'then' st_list[parent]
    {
    	ifst.setGuard($assign_expr.str);
	$ifst.setThenGoto(new GotoStatement(2));
	for (Statement st : $st_list.l) {
	    $ifst.addThenStatement(st);
	}
	$ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
    }
    ;

else_cases [Statement parent, IfStatement ifst] :
    elif_case[parent] (e=else_cases[parent,$elif_case.ifst])?
    {
	$ifst.addElseStatement($elif_case.ifst);
	$elif_case.ifst.setExitGoto(new GotoStatement($elif_case.ifst.getNbElseStatements()+1));
    }
    | 'else' st_list[parent]
    {
    	for (Statement st : $st_list.l) {
	    $ifst.addElseStatement(st);
	}
    }
    ;

elif_case [Statement parent] returns [IfStatement ifst]
@init {
    ifst = new IfStatement();
    // parent is never null
} :
'elif' '(' assign_expr[true,parent] ')' 'then' st_list[parent]
{
    $ifst.setGuard($assign_expr.str);
    $ifst.setThenGoto(new GotoStatement(2));
    for (Statement st : $st_list.l) {
	$ifst.addThenStatement(st);
    }
    $ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
}
;

assign_expr [boolean guard, Statement st] returns [String str]:
    a=and_expr[guard,st]		{ $str = $a.str; }
    ( '||' b=assign_expr[guard,st]	{ try {
					    BigInteger val_a = new BigInteger($a.str);
					    BigInteger val_b = new BigInteger($b.str);
					    if (val_a.signum()!=0 || val_b.signum()!=0) {
						$str = new String("1");
					    }
					    else {
						$str = new String("0");
					    } 
					  }
					  catch(NumberFormatException e) {
					    $str = new String("(" + $a.str + ")||(" + $b.str + ")");
					  } 
					}
    )?
    ;
	   	   
and_expr [boolean guard, Statement st] returns [String str]:	
    a=xor_expr[guard,st]		{ $str = $a.str; }
    ( '&&' b=and_expr[guard,st]		{
					  try {
					      BigInteger val_a = new BigInteger($a.str);
					      BigInteger val_b = new BigInteger($b.str);
					      if (val_a.signum()!=0 && val_b.signum()!=0) {
						  $str = new String("1");
					      }
					      else {
						  $str = new String("0");
					      }
					  }
					  catch(NumberFormatException e) {
					      $str = new String("(" + $a.str + ")&&(" + $b.str + ")");
					  }
					}
    )?
    ;

xor_expr [boolean guard, Statement st] returns [String str]:	
    a=comp_expr[guard,st]		{ $str = $a.str; }
    ( '^' b=xor_expr[guard,st]		{ try {
					      BigInteger val_a = new BigInteger($a.str);
					      BigInteger val_b = new BigInteger($b.str);
					      if ((val_a.signum()==0 && val_b.signum()!=0) || (val_a.signum()!=0 && val_b.signum()==0)) {
						  $str = new String("1");
					      }
					      else {
						  $str = new String("0");
					      }
					  }
					  catch(NumberFormatException e) {
					      $str = new String("(" + $a.str + ")^(" + $b.str + ")");
					  }
				        }
    )?
    ;

        
comp_expr [boolean guard, Statement st] returns [String str]:
a=sum_expr[guard,st]			{ $str = $a.str; }
( '<=' b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")<=(" + $b.str + ")"); }
| '>=' b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")>=(" + $b.str + ")"); }
| '<'  b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")<(" + $b.str + ")"); }
| '>'  b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")>(" + $b.str + ")"); }
| '==' b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")==(" + $b.str + ")"); }
| '!=' b=sum_expr[false, st]		{ $str = new String("(" + $a.str + ")!=(" + $b.str + ")"); }
)?
;

sum_expr [boolean guard, Statement st] returns [String str]:
    a=prod_expr[guard, st]		{   $str = $a.str; }
    ( '+' b=sum_expr[guard, st]		{   if(guard)
						throw new RuntimeException("Sums are not allowed in left members of a guard.");
					    try {
						BigInteger val_a = new BigInteger($a.str);
						BigInteger val_b = new BigInteger($b.str);
						$str = val_a.add(val_b).toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("(" + $a.str + ")+(" + $b.str + ")");
					    }
					}
     | '-' b=sum_expr[guard, st]	{   if(guard)
						throw new RuntimeException("Differences are not allowed in left members of a guard.");
					    try {
						BigInteger val_a = new BigInteger($a.str);
						BigInteger val_b = new BigInteger($b.str);
						$str = val_a.subtract(val_b).toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("(" + $a.str + ")-(" + $b.str + ")");
					    }
					}
    )?
    ;

prod_expr [boolean guard, Statement st] returns [String str]:
    a=unar_expr[guard, st]		{  $str = $a.str; }
    ( '*' b=prod_expr[guard, st]	{  if(guard)
					      throw new RuntimeException("Poducts are not allowed in left members of a guard.");
					    try {
						BigInteger val_a = new BigInteger($a.str);
						BigInteger val_b = new BigInteger($b.str);
						$str = val_a.multiply(val_b).toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("(" + $a.str + ")*(" + $b.str + ")");
					    }
					}
    | '/' b=prod_expr[guard, st]    	{   if(guard)
						throw new RuntimeException("Divisions are not allowed in left members of a guard.");    
					    try {
						BigInteger val_a = new BigInteger($a.str);
						BigInteger val_b = new BigInteger($b.str);
						$str = val_a.divide(val_b).toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("(" + $a.str + ")/(" + $b.str + ")");
					    }
					}
    | '\%' b=prod_expr[guard, st]		{   if(guard)
						throw new RuntimeException("Modulos are not allowed in left members of a guard.");
					    try {
						BigInteger val_a = new BigInteger($a.str);
						BigInteger val_b = new BigInteger($b.str);
						$str = val_a.mod(val_b).toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("(" + $a.str + ")\%(" + $b.str + ")");
					    }
					}
    )?
    ;

unar_expr [boolean guard, Statement st] returns [String str]:
    term[guard, st] 			{   $str = $term.str; }
    | '!' term[guard, st]	    	{   if(guard)
					        throw new RuntimeException("Negations are not allowed in left members of a guard.");
					    try {
						BigInteger val = new BigInteger($term.str);
						if(val.signum()==0) {
						    $str = new String("1");
						}
						else {
						    $str = new String("0");
						} 
					    }
					    catch(NumberFormatException e) {
						$str = new String("!(" + $term.str + ")");
					    }
					}
    | '-' term[guard, st]		{   if(guard)
						throw new RuntimeException("Negations are not allowed in left members of a guard.");
					    try {
						BigInteger val = new BigInteger($term.str);
						$str = val.negate().toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("-(" + $term.str + ")");
					    } 
					}
    ;

term [boolean guard, Statement st] returns [String str]:
    '(' assign_expr[false,st] ')'	{   if (guard)
						throw new RuntimeException("Expressions are not allowed in left members of a guard.");
					    $str = $assign_expr.str; 
					}
    | leaf_term[guard,st]		{   $str = $leaf_term.str; }
    ;

leaf_term  [boolean guard, Statement st] returns [String str]:
    NUMBER 				{   $str = $NUMBER.text; }
    | variable[guard, true, st]		{   $str = $variable.str; }
    ;	 

variable [boolean vprivate, boolean constant, Statement st] returns [String str]:
    IDENTIFIER indices[st]		{   BigDecimal val = prog.getConstantValue($IDENTIFIER.text);
					    if (val != null) {
						if (!constant)
						    throw new RuntimeException("Constant are not allowed in this expression.");
						if ($indices.l.isEmpty()) {
						    $str = val.toBigInteger().toString();
						}
						else {
						    System.err.println("ERROR: on line " + $IDENTIFIER.getLine()  + ", constant " + $IDENTIFIER.text + " cannot be an array.");
						    System.exit(1);
						}
					    }
					    else {
						Variable v = prog.getVariable($IDENTIFIER.text);
						if(v != null) {
						    if (!vprivate && v.getPublicity() < 0)
							throw new RuntimeException("Private variables are not allowed in this expression.");
						    if (v.getPublicity() == 1 && $st != null) {
						      v.setLastUsed($st);
						      $st.addUsedVariable(v);
						    }
						    $str = $IDENTIFIER.text;
						    for (Iterator<String> it = $indices.l.iterator(); it.hasNext();) {
							String s = it.next();
							$str = $str.concat(s);
						    }

						    if (v.getDimensions() < $indices.l.size()) {
							System.err.println("ERROR: on line " + $IDENTIFIER.getLine() + ", " + $str + " is not an array.");
							System.exit(1);
						    }
						    else if (v.getDimensions() > $indices.l.size()) {
							System.err.println("ERROR: on line " + $IDENTIFIER.getLine() + ", " + $str + " is an array.");
							System.exit(1);
						    }
						}
						else {
						    System.err.println("ERROR: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is not declared.");
						    System.exit(1);
						}
					    }
					}
    ;

indices [Statement st] returns [LinkedList<String> l]
    @init{ $l = new LinkedList<String>(); }
    :
    ('[' assign_expr[false,st] ']' 		{ $l.add(new String("[(" + $assign_expr.str + ")]")); }
    )*
    ;
    
    
float_number returns [BigDecimal f]:
n1=NUMBER '.' n2=NUMBER 	{ $f = new BigDecimal($n1.text + "." + $n2.text); }
;

    	 
INT :		'int'('a'..'z'|'A'..'Z'|'0'..'9'|'_')+ ;
IDENTIFIER  :   ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'0'..'9'|'_')* ;
NUMBER :	'0'..'9'+ ;
// FLOAT :		'0''.'('0'..'9')+ ;
NEWLINE :	'\n' { skip(); } ;
SPACES :	(' '|'\t'|'\r')+ { skip(); } ;
COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' { $channel=HIDDEN; } ;
