/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Louis-Marie Traonouez <louis-marie.traonouez@inria.fr>
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Iterator;
import java.util.LinkedList;

/** An abstract for QUAIL variables.
 *  Variables can be public or private, observable or secret, integer or boolean, or multi-dimensional arrays of variables.
 *  @author ltraonou
 */
public abstract class Variable {

    protected String name;
    
    /** Number of bits */
    protected int intType;
    /** Minimum value (always zero in current implementations) */
    protected BigInteger minInt;
    /** Maximum value (2^intType-1)  */
    protected BigInteger maxInt;
    /** Dimensions of the array. If none the variable is not an array. */
    protected LinkedList<ArrayDimension> arrays;
    /** Last statement that uses the variable. */
    protected Statement lastUsed; 
    
    public void setName(String n) {
    	this.name = n;
    }
    
    public String getName() {
    	return this.name;
    }
     
    public abstract void setPublicity(boolean b);
    
    // 1 for public
    // 2 for observable
    // -1 for private
    // -2 for secret
    public abstract int getPublicity();
    
    public void setLastUsed(Statement st) {
    	lastUsed = st;
    }

    public Statement getLastUsed() {
    	return lastUsed;
    }
    
    public int getIntType() {
        return intType;
    }
     
    public int getDimensions() {
        return arrays.size();
    }
    
    public ArrayDimension getArrayDimension(int i) {
        if (i < arrays.size() && i >= 0) {
            return arrays.get(i);
        }
        else {
            return null;
        }
    }
    
    public int getNbValues() {
        int nb = 1;
        for (Iterator<ArrayDimension> it = arrays.iterator(); it.hasNext();) {
            ArrayDimension arrayDimension = it.next();
            nb = nb * arrayDimension.getNbIndices();
        }
        return nb;
    }
    
    public void addDimension(ArrayDimension dim) {
        arrays.addFirst(dim);
    }
    
    public LinkedList<String> generateIndices() {
        LinkedList<String> l = new LinkedList<String>();
        if (getDimensions() > 0) {
            l.add("");
            for (Iterator<ArrayDimension> ar = arrays.iterator(); ar.hasNext();) {
                ArrayDimension dim = ar.next();
                LinkedList<String> newl = new LinkedList<String>();
                for (Iterator<String> it = l.iterator(); it.hasNext();) {
                    String s = it.next();
                    for (int i = dim.getMinIndex(); i <= dim.getMaxIndex(); i++) {
                        newl.add(s.concat("["+i+"]"));
                    }
                }
                l = newl;
            }
        }
        return l;
    }
    
    public LinkedList<FreeStatement> freeVariable() {
    	LinkedList<FreeStatement> ret = new LinkedList<FreeStatement>();
        if (getDimensions() > 0) {
	    	LinkedList<String> indices = generateIndices();
	    	for (String index : indices) {
	    		ret.add(new FreeStatement(name+index));
	    	}
        }
        else {
        	ret.add(new FreeStatement(name));
        }
    	return ret;
    }
    
    public abstract int print(PrintWriter out);
}
