/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/

package quail;

import java.io.FileNotFoundException;
import java.math.BigDecimal;
import java.util.ArrayList;

/**
 *
 * @author fbio
 */
public class Analyzer {
    
   public Analyzer(){
       
   }

   public static void analyze(String filename, String[] printMc, boolean run_hiding, boolean run_rate) {
        ///////////////////////////////////////////////////////
        // read attacker
        
        /* PROTOTYPE: we assume that the attacker has uniform prior
         * information, can observe only initial and absorbing 
         * states and can discriminate states with different 
         * values of the public variables.
         * 
         */
        
        ///////////////////////////////////////////////////////
        
        ///////////////////////////////////////////////////////
        // read WHILE code of program
        Logger logger = Logger.getInstance();
        Contexter contexter = Contexter.getInstance();  
        
        long time = System.nanoTime();
        CodeReader reader = new CodeReader(filename);
        
        ArrayList<String> codelines = null;
        try{
            codelines = reader.readLines();
        }catch (FileNotFoundException fnf)
        {
            log("Error while processing input file: file " + filename + " not found",1);
            
        }
        boolean randomized = reader.isRandomized();
        if (codelines==null){
            log("ERROR: No code lines read? Terminating",1);
            System.exit(1);
        }
        log(codelines.toString(),3);
        CodeParser parser = new CodeParser();
        MarkovStateSet inputset = parser.parseCode(codelines);
    
        if (inputset == null){
            log("Code parsing failed; terminating.",1);
            return;
        }
        log ("Parsing terminated at time: "+(System.nanoTime()-time)/1000000+" milliseconds",1);
        ///////////////////////////////////////////////////////
        
        ///////////////////////////////////////////////////////
        // build Markov chain
        
        /*
         * Example Markov chain: mc for h XOR random(p)
         * 
         * double p = 0.5;
        double prinf = 0.5;
        String[] states = new String[]{"start","s1","s2","s3","s4"};
        double[] initprobs = new double[]{1.0,0.0,0.0,0.0,0.0};
        double[][] trans = new double[5][5];
        for (int i=0; i<5; i++){
            for (int j=0; j<5; j++){
                trans[i][j] = 0.0;
            }
        }
        
        trans[0][1] = p*prinf;
        trans[0][2] = (1-p)*(1-prinf);
        trans[0][3] = (1-p)*prinf;   
        trans[0][4] = p*(1-prinf);       
                
        trans[1][1] = 1.0;
        trans[2][2] = 1.0;
        trans[3][3] = 1.0;
        trans[4][4] = 1.0;
        
        MarkovChain mc = new MarkovChain(states,initprobs,trans);
        mc.printNice();
        
        MarkovStatePublicVariable l0 = new MarkovStatePublicVariable("l",0,1);
        MarkovStatePublicVariable l1 = new MarkovStatePublicVariable("l",1,1);
        MarkovStatePrivateVariable h0 = new MarkovStatePrivateVariable("h","[0,0]",1);
        MarkovStatePrivateVariable h1 = new MarkovStatePrivateVariable("h","[1,1]",1);
        
        MarkovChainState s1 = mc.getStateByName("s1");
        s1.addVariable(l0);
        s1.addVariable(h0);
        
        MarkovChainState s2 = mc.getStateByName("s2");
        s2.addVariable(l0);
        s2.addVariable(h1);
        
        MarkovChainState s3 = mc.getStateByName("s3");
        s3.addVariable(l1);
        s3.addVariable(h0);
        
        MarkovChainState s4 = mc.getStateByName("s4");
        s4.addVariable(l1);
        s4.addVariable(h1);
         */
        
        MarkovChain mc = new MarkovChain(inputset);
        if (logger.getVerbosity()>=3){
            mc.printNice(3);
        }
        if (printMc[1]!=null) {
            log("Printing Markov chain to file "+printMc[1],1);
            mc.printGraphToFile(printMc[1],printMc[0]);
        }
        
    //    double entropy_mc = mc.calculateEntropy();
    //    log("Its entropy is "+entropy_mc);
        
        
        ///////////////////////////////////////////////////////
        
        ///////////////////////////////////////////////////////
        // Hide the non observable states
        
        if (run_hiding){        
            //create a list of states to hide
            MarkovStateSet states = mc.getStates();
            MarkovStateSet tohide = new MarkovStateSet();
           for (MarkovChainState s : states) {
               if (s.getInitialProbability().compareTo(contexter.getDecimalOne())==0){
                   continue;
               }

               //if it is an absorbing state, don't add it
               if ((mc.getTransitionProbabilityFromTo(s.getName(), s.getName())).compareTo(contexter.getDecimalOne())==0){
                   continue;
               }
               //else add it to the states to be hidden
               tohide.add(s);
           }
            if (logger.getVerbosity()>=3){
                log("Hiding the states "+tohide.toString(),3);
            }

            mc.hideStates(tohide);
            log("",2);
            log ("Hiding terminated at time: "+(System.nanoTime()-time)/1000000+" milliseconds",1);
            log("The observable reduction of the Markov Chain has "+mc.getStates().size()+" states.",2);
            if (logger.getVerbosity()>=3){
                mc.printNice(3);
            }
            if (printMc[2]!=null) {
                log("Printing hidden Markov chain to file "+printMc[2],1);
                mc.printGraphToFile(printMc[2],printMc[0]);
            }
        }
  //      entropy_mc = mc.calculateEntropy();
  //      log("Its entropy is "+entropy_mc);
        
         ///////////////////////////////////////////////////////
        
        ///////////////////////////////////////////////////////
        // build the three discrimination relations
        log("Building the identity discrimination relation...",3);
        DiscriminationRelation identity = new DiscriminationRelation(mc.getStates());
        if (logger.getVerbosity() >=3){
            log("Printing the identity discrimination relation:",3);
            identity.printNice(3);
        }
        
        
        log("",2);
        ArrayList<String> obsvars = mc.getObservableVariables(); 
        log("The observable variables are: "+obsvars,2);
        if (obsvars.isEmpty()){
            log("ERROR: no observable variable defined; the leakage is 0. Terminating.",1);
            return;
        }
        log("Computing the observer's discrimination relation",2);
        DiscriminationRelation disc_l = identity.DiscriminateBy(obsvars.get(0));
        for (int i=1; i<obsvars.size(); i++){
            disc_l = disc_l.intersectBy(obsvars.get(i));
        }
        if (logger.getVerbosity()>=3){
            log("Printing the observer's discrimination relation:",3);
            disc_l.printNice(3);
        }
        
        log("Computing the observer's quotient...",2);
        MarkovChain quotient_l = mc.computeQuotient(disc_l);
        if (logger.getVerbosity()>=3){
            log("Printing the observer's quotient:",3);
            quotient_l.printNice(3);
        }
        if (printMc[3]!=null) {
            log("Printing observer's quotient Markov chain to file "+printMc[3],1);
            quotient_l.printGraphToFile(printMc[3],printMc[0]);
        }
        BigDecimal entropy_l;
        if (run_rate==false){
            log("Computing the entropy of the observer's quotient...",2);
            entropy_l = quotient_l.calculateEntropy();
            log("The entropy of the observer's quotient is "+entropy_l,2);
        }else{
            log("Computing the entropy rate of the observer's quotient...",2);
            entropy_l = quotient_l.calculateEntropyRate();
            log("The entropy rate of the observer's quotient is "+entropy_l,2);
        }
        log ("Entropy of observer computed at time: "+(System.nanoTime()-time)/1000000+" milliseconds",1);
        
        log("",2);

        
        // uncomment next line to force QUAIL to compute all 3 quotients, even when not needed
        //randomized = true;
        
        
        BigDecimal leakage;
        if (randomized){
            log("The program is randomized; computing secret and joint quotient",3);
            ArrayList<String> secvars = mc.getSecretVariables(); 
            log("The secret variables are: "+secvars,2);
            if (secvars.isEmpty()){
                log("ERROR: no secret variable defined; the leakage is 0. Terminating.",1);
                return;
            }
            log("Computing the secret's discrimination relation...",2);
            DiscriminationRelation disc_h = identity.DiscriminateBy(secvars.get(0));
            for (int i=1; i<secvars.size(); i++){
                disc_h = disc_h.intersectBy(secvars.get(i));
            }
            if (logger.getVerbosity()>=3){
                log("Printing the secret's discrimination relation:",3);
                disc_h.printNice(3);
            }
            
            log("Computing the secret's quotient...",3);
            MarkovChain quotient_h = mc.computeQuotient(disc_h);
            if (logger.getVerbosity()>=3){
                log("Printing the secret's quotient:",3);
                quotient_h.printNice(3);
            }
            if (printMc[4]!=null) {
                log("Printing secret's quotient Markov chain to file "+printMc[4],1);
                quotient_h.printGraphToFile(printMc[4],printMc[0]);
            }
            BigDecimal entropy_h;
            if (run_rate==false){
                log("Computing the entropy of the secret's quotient...",2);
                entropy_h = quotient_h.calculateEntropy();
                log("The entropy of the secret's quotient is "+entropy_h,2);
            }else{
                log("Computing the entropy rate of the secret's quotient...",2);
                entropy_h = quotient_h.calculateEntropyRate();
                log("The entropy rate of the secret's quotient is "+entropy_h,2);
            }

            
            log ("Entropy of secret computed at time: "+(System.nanoTime()-time)/1000000+" milliseconds",1);
            log("",2);
            log("Computing the joint discrimination relation...",2);
            DiscriminationRelation disc_lh = disc_l.intersectBy(secvars.get(0));
            for (int i=1; i<secvars.size(); i++){
                disc_lh = disc_lh.intersectBy(secvars.get(i));
            }
            
            if (logger.getVerbosity()>=3) {
                log("Printing the joint discrimination relation:",3);
                disc_lh.printNice(3);
            }
            log("Computing the joint quotient...",2);
            MarkovChain quotient_lh = mc.computeQuotient(disc_lh);
            if (logger.getVerbosity()>=3){
                log("Printing the joint quotient:",3);
                quotient_lh.printNice(3);
            }
            if (printMc[5]!=null) {
               log("Printing joint quotient Markov chain to file "+printMc[5],1);
               quotient_lh.printGraphToFile(printMc[5],printMc[0]);
            }
            
            BigDecimal entropy_lh;
            if (run_rate==false){
                log("Computing the entropy of the joint quotient...",2);
                entropy_lh = quotient_lh.calculateEntropy();
                log("The entropy of the joint quotient is "+entropy_lh,2);
            }else{
                log("Computing the entropy rate of the joint quotient...",2);
                entropy_lh = quotient_lh.calculateEntropyRate();
                log("The entropy rate of the joint quotient is "+entropy_lh,2);
            }

            ///////////////////////////////////////////////////////
            // compute the leakage

            leakage = entropy_l.add(entropy_h).subtract(entropy_lh);
        }else{
            log("The program is deterministic; the leakage is the observer's entropy",2);
            leakage = entropy_l;
        }
        
        
        log("",2);
        log("===========================",2);
        if (logger.getVerbosity()>0){
            if (run_rate==false){
                log("The leakage is "+leakage+" bits with a precision of "+leakage.precision()+" digits",1);
            }else{
                log("The leakage rate is "+leakage+" bits with a precision of "+leakage.precision()+" digits",1);
            }
            
        }else{
            log(leakage,0);
        }
        
        log("===========================",2);
        
        log ("Expired time: "+(System.nanoTime()-time)/1000000+" milliseconds",1);
        ///////////////////////////////////////////////////////
    }

   private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }

  

}