/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.text.ParseException;

/**
 *
 * @author fbio
 */
public class CodeParser {
    

  private final HashSet<Transition> transset;
  private final MarkovStateSet stateset;
  private final MarkovStateSet toprocess;
  private long statecount;
  MarkovChainState currstate;
  private boolean parsingerror = false;
  private static final Contexter contexter = Contexter.getInstance();
  private static final Logger logger = Logger.getInstance();
  
 
  
        
  
  
  private static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }
    
  public CodeParser(){
    statecount = 1;
    stateset = new MarkovStateSet();
    toprocess = new MarkovStateSet();
    transset = new HashSet<>();
  
    
    MarkovChainState startstate = new MarkovChainState("start",contexter.getDecimalOne());
    MarkovStatePublicVariable pc = new MarkovStatePublicVariable("pc",BigInteger.ZERO,Integer.valueOf(32));
    startstate.addVariable(pc);
    
    log("Creating start state",3);
    stateset.add(startstate);
    VariableList vars = startstate.getCloneOfVariables();
    vars.setPublicVariable("pc", 1);
    currstate = this.getNextState(vars);
    
    createTransition(startstate.getName(),currstate.getName(),contexter.getDecimalOne());
    
  }
  
  public MarkovStateSet parseCode(ArrayList<String> linelist){
    log("Processing "+(linelist.size()-1)+" code lines",3);
    Iterator<MarkovChainState> stateiter = toprocess.iterator();
    while (stateiter.hasNext()){
          currstate =stateiter.next();
          BigInteger pc = currstate.getPublicVariableValue("pc");
          if (pc.compareTo(BigInteger.valueOf(-1)) ==0){
              log("ERROR: found state with no program counter information; exiting parser",1);
              parsingerror = true;
              break;
          }
          
          String line = linelist.get(pc.intValue()); 
          log("",4);
          if (logger.getVerbosity()>=4){
              log("Processing line "+pc+": "+line+" in state "+currstate.printNice(),4);
          }
          
          if (line.startsWith("private")){
              
              processPrivateLine(line,pc,false);
          }else if (line.startsWith("public")){
              processPublicLine(line,pc,false);
          }else if (line.startsWith("secret")){
              
              processPrivateLine(line,pc,true);
          }else if (line.startsWith("observable")){
              processPublicLine(line,pc,true);
          }else if (line.startsWith("assign")){
              processAssignLine(line,pc);
          }else if (line.startsWith("skip")){
              processSkipLine(line,pc);
          }else if (line.startsWith("free")){
              processFreeLine(line,pc);
          }else if (line.startsWith("random")){
              processRandomLine(line,pc);
          }else if (line.startsWith("goto")){
              processGotoLine(line,pc);
          }else if (line.startsWith("return")){
              processReturnLine(line,pc);
          }else if (line.startsWith("if")){
              processIfLine(line,linelist.get(pc.add(BigInteger.ONE).intValue()),linelist.get(pc.add(new BigInteger("2")).intValue()),pc);
          }else {
              log("ERROR: line "+pc.toString()+" unrecognized, this should never happen; exiting parser",1);
              parsingerror = true;
              break;
          }
          if (parsingerror){
              break;
          }
          stateiter = toprocess.iterator();
          

      }
    
    if (parsingerror){
        return null;
    }
    log("Built "+stateset.size()+" states and "+transset.size()+" transitions",2);
    log("Embedding transitions in the states...",4);
    Iterator<Transition> transiter = transset.iterator();
    while (transiter.hasNext()){
        Transition trans =transiter.next();
        MarkovChainState source = stateset.getStateByName(trans.getSource());
        MarkovChainState dest = stateset.getStateByName(trans.getDest());
        source.getSuccessors().add(dest);
        dest.getPredecessors().add(source);
        source.getTransitionsFrom().add(trans);
    }
    return stateset;
    
  }
  
  protected void processReturnLine(String aLine, BigInteger pc){
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_return.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_return.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
      
    createTransition(currstate.getName(),currstate.getName(),contexter.getDecimalOne());
    stateset.add(currstate);
    toprocess.remove(currstate);
  }
  
  protected void processPrivateLine(String aLine, BigInteger pc, boolean isSecret){

      
    String varname;
    Integer varlength;
    String varinterval;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    
    Pattern pattern;
    if (isSecret){
        pattern = QuailPatterns.pattern_secret;
    }else{
        pattern = QuailPatterns.pattern_private;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      BigInteger interend = BigInteger.valueOf(2).pow(varlength.intValue()).subtract(BigInteger.ONE);
      varinterval="[0,"+interend+"]";
    }else{
        if (isSecret){
        pattern = QuailPatterns.pattern_secret_with_assignment ;
    }else{
        pattern = QuailPatterns.pattern_private_with_assignment ;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable length: " + matcher.group(1),4);
          log("Variable name: " + matcher.group(2),4);
          log("Assigned interval: "+matcher.group(3),4);
          varlength = Integer.valueOf(matcher.group(1));
          varname = matcher.group(2);
          varinterval=matcher.group(3);
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
            this.toprocess.remove(currstate);
            parsingerror = true;
            return;
        }
        
    }
    
    
    
    
    MarkovStatePrivateVariable var = new MarkovStatePrivateVariable(varname,varinterval,varlength);
    var.setSecret(isSecret);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isSecret){
            log("Added secret variable "+var.toString()+" with "+var.getIntervalSize()+" possible values "+varinterval+" to state "+currstate.getName(),3);        
        }else{
            log("Added private variable "+var.toString()+" with "+var.getIntervalSize()+" possible values "+varinterval+" to state "+currstate.getName(),3);
        }
    }
    
    
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    //
    VariableList vars;
      vars = currstate.getCloneOfVariables();
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    vars.setPublicVariable("pc", newpc);
    MarkovChainState newstate = getNextState(vars);
                
   
    createTransition(currstate.getName(),newstate.getName(),contexter.getDecimalOne());
   
 
  }
  
   protected void processPublicLine(String aLine, BigInteger pc, boolean isObservable){
    if (isObservable){
        log("Parsing observable line: "+aLine,4);
    }else{
        log("Parsing public line: "+aLine,4);
    }
       
       
    //Scanner scanner = new Scanner(aLine);
    String varname;
    Integer varlength;
    BigInteger varvalue;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    /*
    scanner.findInLine("public int[\d+] \w+;");
    MatchResult result = scanner.match();*/
    Pattern pattern;
    if (isObservable){
        pattern = QuailPatterns.pattern_observable ;
    }else{
        pattern = QuailPatterns.pattern_public ;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      varvalue = BigInteger.ZERO;
    
    }else{
        if (isObservable){
            pattern = QuailPatterns.pattern_observable_with_assignment ;
        }else{
            pattern = QuailPatterns.pattern_public_with_assignment ;
        }
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+pattern.pattern(),4);
        }
        
        matcher = pattern.matcher(aLine );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable length: " + matcher.group(1),4);
          log("Variable name: " + matcher.group(2),4);
          log("Assigned value: "+matcher.group(3),4);
          varlength = Integer.valueOf(matcher.group(1));
          varname = matcher.group(2);
          try{
          varvalue = evaluateExpression(matcher.group(3),currstate.getVariables());
          } catch (ParseException e) {
              log("ERROR: unable to evaluate expression "+matcher.group(3)+" in state "+currstate.printNice(),1);
              toprocess.remove(currstate);
              parsingerror = true;
              return;
          }
    
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
            this.toprocess.remove(currstate);
            parsingerror = true;
            return;
        }
    }
    

    MarkovStatePublicVariable var = new MarkovStatePublicVariable(varname,varvalue,varlength);
    var.setObservable(isObservable);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isObservable){
            log("Added observable variable "+var.toString()+" to state "+currstate.getName(),3);
        }else{
            log("Added public variable "+var.toString()+" to state "+currstate.getName(),3);
        }
    }
    
    
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    //
    VariableList vars =currstate.getCloneOfVariables();
    
   // log("Old vars: "+vars.getIdentifier());
    BigInteger newpc = pc.add(BigInteger.ONE);
  //  log("New pc: "+newpc);
    vars.setPublicVariable("pc", newpc);
  //  log("New vars: "+vars.getIdentifier());
    
    MarkovChainState newstate = getNextState(vars);
                
   
    createTransition(currstate.getName(),newstate.getName(),contexter.getDecimalOne());
  
   
 
  }
   
   protected void processGotoLine(String aLine,BigInteger pc){
   
    BigInteger gotoline;
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_goto.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_goto.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Goto line destination: " + matcher.group(1),4);
      gotoline = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
   
    log("Adding goto transition from state "+currstate.getName()+" to line "+gotoline,3);
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    
    VariableList vars =currstate.getCloneOfVariables();    

    
    BigInteger newpc = gotoline;
   
    vars.setPublicVariable("pc", newpc);
 
    MarkovChainState newstate = getNextState(vars);
                

    createTransition(currstate.getName(),newstate.getName(),contexter.getDecimalOne());
   
 
  }
   
   protected void processSkipLine(String aLine,BigInteger pc){
   
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_skip.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_skip.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    log("Adding skip transition in state "+currstate.getName(),3);
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    
    VariableList vars =currstate.getCloneOfVariables();
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    vars.setPublicVariable("pc", newpc);
    MarkovChainState newstate = getNextState(vars);
                
 
    createTransition(currstate.getName(),newstate.getName(),contexter.getDecimalOne());
 
 
  }
   
   protected void processRandomLine(String line, BigInteger pc) {
        String varname;
        
        
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+QuailPatterns.pattern_randombit.pattern(),4);
        }
        
        Matcher matcher = QuailPatterns.pattern_randombit.matcher(line );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Probability of being 0: " + matcher.group(2),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());
          BigDecimal probofzero = new BigDecimal(matcher.group(2),contexter.getContext());
          
          if (probofzero.compareTo(contexter.getDecimalOne())>0){
                log("Found a probability greater than 1; using 1 instead",4);
                probofzero = contexter.getDecimalOne();
            } else if (probofzero.compareTo(contexter.getDecimalZero())<0){
                log("Found a probability lesser than 0; using 0 instead",4);
                probofzero = contexter.getDecimalZero();
            }

            VariableList varszero =currstate.getCloneOfVariables();
            VariableList varsone =currstate.getCloneOfVariables();

            log("Assigned to public variable "+varname+" value 0 with probability "+probofzero+" in state "+currstate.getName(),3);
            stateset.add(currstate);
            toprocess.remove(currstate);

            BigInteger newpc = pc.add(BigInteger.ONE);
            varszero.setPublicVariable("pc", newpc);
            varsone.setPublicVariable("pc", newpc);

            varszero.setPublicVariable(varname, 0);
            varsone.setPublicVariable(varname, 1);

            if (probofzero.compareTo(contexter.getDecimalZero())>0){
               MarkovChainState zerostate = getNextState(varszero);
               createTransition(currstate.getName(),zerostate.getName(),probofzero);
            }

            if (probofzero.compareTo(contexter.getDecimalOne())<0){
                MarkovChainState onestate = getNextState(varsone);
                createTransition(currstate.getName(),onestate.getName(),contexter.getDecimalOne().subtract(probofzero));
            }

          
          
        }else{
            if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+QuailPatterns.pattern_random.pattern(),4);
            }

            matcher = QuailPatterns.pattern_random.matcher(line );
            if ( matcher.find() ) {
                log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Lower bound: " + matcher.group(2),4);
          log("Upper bound: " + matcher.group(3),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());

        BigInteger lowbound = new BigInteger("0");
        try{
            lowbound = evaluateExpression(matcher.group(2),currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }

 
        BigInteger uppbound = new BigInteger("0");
        try{
            uppbound = evaluateExpression(matcher.group(3),currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }
           
          if (lowbound.compareTo(uppbound)>0){
              log("ERROR: lower bound greater than upper bound?",2);
              this.toprocess.remove(currstate);
              parsingerror = true;
              return;
          }
          
          BigInteger diff = uppbound.subtract(lowbound).add(BigInteger.ONE);
          if (diff.compareTo(BigInteger.ONE)==0){
            VariableList vars =currstate.getCloneOfVariables();

            log("Assigned to public variable "+varname+" value "+uppbound.toString()+" with probability 1 in state "+currstate.getName(),3);
            stateset.add(currstate);
            toprocess.remove(currstate);

            BigInteger newpc = pc.add(BigInteger.ONE);
            vars.setPublicVariable("pc", newpc);
           
            vars.setPublicVariable(varname, uppbound);
            
            MarkovChainState nextstate = getNextState(vars);
            this.createTransition(currstate.getName(),nextstate.getName(),contexter.getDecimalOne());
            
          }else{
              BigDecimal decdiff = new BigDecimal(diff);
              BigDecimal prob = contexter.getDecimalOne().divide(decdiff, contexter.getContext());
              while (diff.compareTo(BigInteger.ZERO)>0){
                  VariableList vars =currstate.getCloneOfVariables();

                  
                  BigInteger newpc = pc.add(BigInteger.ONE);
                  vars.setPublicVariable("pc", newpc);

                  vars.setPublicVariable(varname, lowbound.add(diff).subtract(BigInteger.ONE));

                  MarkovChainState nextstate = getNextState(vars);
                  this.createTransition(currstate.getName(),nextstate.getName(),prob);
                  diff = diff.subtract(BigInteger.ONE);
              }
              log("Assigned to public variable "+varname+" a random value from "+lowbound.toString()+" to "+uppbound.toString()+" with uniform probability "+prob.toString()+" in state "+currstate.getName(),3);
              stateset.add(currstate);
              toprocess.remove(currstate);

          }
        
            }else{
                log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);

                this.toprocess.remove(currstate);
                parsingerror = true;
            }
                
        }
        
        
        
        
                
        
    }
   
   protected void processAssignLine(String aLine, BigInteger pc){
   
  
    String varname;
    String varexpression;
    BigInteger varvalue;
    
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_assign.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_assign.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Variable expression: " + matcher.group(2),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      varexpression = matcher.group(2);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    VariableList vars =currstate.getCloneOfVariables();
    try{
        varvalue = evaluateExpression(varexpression,vars);
    }catch (ParseException e){
        log("ERROR: unable to evaluate expression "+varexpression+" in state "+currstate.printNice(),1);
        toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    
    log("Assigned to public or observable variable "+varname+" value "+varvalue+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    vars.setPublicVariable("pc", newpc);
    vars.setPublicVariable(varname, varvalue);
    
    // VERY EXPERIMENTAL: build a new state only if the variable is observable; if it's public just change its
    // value in the current state? 
    MarkovStatePublicVariable thevar =(MarkovStatePublicVariable) vars.getVariable(varname);
    
    if (thevar.isObservable()){
        stateset.add(currstate);
        toprocess.remove(currstate);
    
        MarkovChainState newstate = getNextState(vars);
        
        createTransition(currstate.getName(),newstate.getName(),Contexter.getInstance().getDecimalOne());
    }else{
        currstate.setVariables(vars);
    }
    
 
  }
   
   private BigInteger evaluateExpression(String expression, VariableList vars) throws ParseException{
       String expr = expression;
       boolean found = false;
       while (found == false){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<expr.length();i++){
               if (expr.charAt(i)==')'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis the expression is a constant; return its value
           if (clospos==-1){
               return this.evaluateInnerExpression(expr, vars);
           
           }else{
               //else find the last opened parenthesis before the closed one
               String smallexpr = expr.substring(0, clospos+1);
               log ("Found closed parenthesis: "+smallexpr,4);
               int openpos = smallexpr.lastIndexOf('(');
               smallexpr = smallexpr.substring(openpos);
               log("Found inner expression "+smallexpr,4);
               // found an inner expression; evaluate it (without the parenthesis)
               BigInteger innerval = new BigInteger("0");
               try{
                   innerval = this.evaluateInnerExpression(smallexpr.substring(1, smallexpr.length()-1), vars);
               }catch (ParseException e){
                   throw e;
               }
               
               String smallexprval = ""+innerval;
               expr = expr.replace(smallexpr, smallexprval);
               log("Replaced inner expression "+smallexpr+" with its value "+smallexprval+" in the string; now the string to evaluate is "+expr,4);
           
           }
       }
       
       
       return BigInteger.valueOf(-1);
   }

    private BigInteger evaluateInnerExpression (String expression, VariableList vars)throws ParseException {
        String expr = expression;
        log("Evaluating inner expression "+expr+" with variables "+vars.toString(),4);
        if (expr.startsWith("(")){
            expr = expr.substring(1,expr.lastIndexOf(")"));
            log("Removing parenthesis; the expression is now "+expr,4);
        }
        if (logger.getVerbosity()>=4){
            log("The regular expression for a constant is "+QuailPatterns.pattern_expression_constant.pattern(),4);
        }
        
        Matcher matcher = QuailPatterns.pattern_expression_constant.matcher(expr );
        if ( matcher.matches() ) {
            log("It is the constant " + matcher.group(1)+"; returning it.",4);
            return new BigInteger(matcher.group(1));
        }else{
            log("It is not a constant, checking if it is a variable",4);
            if (logger.getVerbosity()>=4){
                log("The regular expression for a variable is "+QuailPatterns.pattern_expression_variable.pattern(),4);
            }
            
            matcher = QuailPatterns.pattern_expression_variable.matcher(expr );
            if ( matcher.matches() ) {
                log("It is the variable " + matcher.group(1)+"; checking its value...",4);
                String varname = this.getVariableName(matcher.group(1), vars);
                MarkovStateVariable var = vars.getVariable(varname);
                if (var==null){
                    log("ERROR: variable "+varname+" not found.",4);
                    throw new ParseException("ERROR: variable "+varname+" not found.",0);
                }else if (var.isPrivate()){
                    log("ERROR: variable "+varname+" is private and thus has no defined value.",4);
                    throw new ParseException("ERROR: variable "+varname+" is private and thus has no defined value.",0);
                }else{
                    BigInteger val = ((MarkovStatePublicVariable)var).getVal();
                    log("The value is "+val+"; returning it.",4);
                    return val;
                }
                
                
            }else{
                log("It is not a variable, checking if it is an operation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for an operation is "+QuailPatterns.pattern_expression_operation.pattern(),4);
                }
                
                matcher = QuailPatterns.pattern_expression_operation.matcher(expr );
                if ( matcher.matches() ) {
                    String op1 = matcher.group(1);
                    String operator = matcher.group(2);
                    String op2 = matcher.group(3);
                    log("It is the operation "+op1+" "+operator+" "+op2+"; evaluating it...",4);
                    switch (operator) {
                        case "+":
                        {
                            log("It is a sum; evaluating the operands",4);
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result = val1.add(val2);
                            log("The sum result is "+result,4);
                            return result;
                        }
                        case "-":
                        {
                            log("It is a subtraction; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result = val1.subtract(val2);
                            log("The subtraction result is "+result,4);
                            return result;
                        }
                        case "/":
                        {
                            log("It is a division; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result = val1.divide(val2);
                            log("The division result is "+result,4);
                            return result;
                        }
                        case "*":
                        {
                            log("It is a product; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result = val1.multiply(val2);
                            log("The product result is "+result,4);
                            return result;
                        }
                        case "%":
                        {
                            log("It is a modulo; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result = val1.mod(val2);
                            log("The modulo result is "+result,4);
                            return result;
                        }
                        case "&&":
                        {
                            log("It is a logical AND; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result;
                            if (val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0){
                                result = BigInteger.ZERO;
                            }else{
                                result = BigInteger.ONE;
                            }
                            log("The AND result is "+result,4);
                            return result;
                        }
                        case "||":
                        {
                            log("It is a logical OR; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result;
                            if (val1.compareTo(BigInteger.ZERO)==0 || val2.compareTo(BigInteger.ZERO)==0){
                                result = BigInteger.ZERO;
                            }else{
                                result = BigInteger.ONE;
                            }
                            log("The OR result is "+result,4);
                            return result;
                        }
                        case "^":
                        {
                            log("It is a logical XOR; evaluating the operands",4);
                            
                            BigInteger val1 = new BigInteger(op1);
                            BigInteger val2 = new BigInteger(op2);
                            BigInteger result;
                            if ((val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0) || (val1.compareTo(BigInteger.ZERO)!=0 && val2.compareTo(BigInteger.ZERO)!=0)){
                                result = BigInteger.ZERO;
                            }else{
                                result = BigInteger.ONE;
                            }
                            log("The XOR result is "+result,4);
                            return result;
                        }
                    }
                    log("ERROR: operator not recognized.",1);
                    throw new ParseException("ERROR: operator not recognized;",0);
                }else {
                 log("It is not an operation, checking if it is a negation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for a negation is "+QuailPatterns.pattern_negated_expression.pattern(),4);
                
                }
                matcher = QuailPatterns.pattern_negated_expression.matcher(expr );
                if ( matcher.matches() ) {
                    String negexpr = matcher.group(1);
                   
                    log("It is the negated operation "+negexpr+"; negating it...",4);
                    BigInteger negval = new BigInteger(negexpr.substring(1));
                    
                    if (negval.compareTo(BigInteger.ZERO)==0){
                        negval = BigInteger.ONE;
                    }else{
                        negval = BigInteger.ZERO;
                    }
                    return negval;
                }else {
                    log("ERROR: string "+expr+" does not match any known inner expression.",1);
                    throw new ParseException("ERROR: string "+expr+" does not match any known inner expression.",0);
                }
            }
            }


        }
    }
   
    protected void processIfLine(String ifline, String thenline, String elseline, BigInteger pc){
   
    
    String guard;
    BigInteger gotothen,gotoelse;
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_if.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_if.matcher(ifline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("If guard: " + matcher.group(1),4);
      guard = matcher.group(1);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_then.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_then.matcher(thenline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Then destination: " + matcher.group(1),4);
      gotothen = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE) +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_else.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_else.matcher(elseline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Else destination: " + matcher.group(1),4);
      gotoelse = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE).add(BigInteger.ONE) +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    
    log("Added if statement with guard "+guard+", then line "+gotothen+" and else line "+gotoelse+" to state "+currstate.getName(),3);
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    String varname;
    String operator;
    String expression;
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_single_guard.pattern(),4);
    }
    
    matcher = QuailPatterns.pattern_guard.matcher(guard );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Operator: " + matcher.group(2),4);
      log("Guard expression: "+matcher.group(3),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      operator = matcher.group(2);
      expression = matcher.group(3);
    }else{
        matcher = QuailPatterns.pattern_single_guard.matcher(guard);
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Operator: " + matcher.group(2),4);
          log("Guard expression: "+matcher.group(3),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());
          operator = matcher.group(2);
          expression = matcher.group(3);
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
            this.toprocess.remove(currstate);
            parsingerror = true;
            return;
       }
    }
   
    // evaluate the guard expression
        BigInteger expval = new BigInteger("0");
        try{
            expval = this.evaluateExpression(expression, currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
        }
        log("The expression evaluates to "+expval,4);
    
   boolean secretdependent = true;
   BigInteger argval = BigInteger.ZERO;
   log("The regular expression for a constant is "+QuailPatterns.pattern_expression_constant.pattern(),4);
   matcher = QuailPatterns.pattern_expression_constant.matcher(varname);
   if ( matcher.matches() ) {
       log("The argument is the constant " + matcher.group(1),4);
       argval = new BigInteger(matcher.group(1));
       secretdependent = false;
   }else{
    
        // obtain the variable
        MarkovStateVariable var = currstate.getVariables().getVariable(varname);
            
        // if the variable is private, the if is secret-dependent; if not take the value
        if (var.isPrivate()==false){
            MarkovStatePublicVariable pubvar = (MarkovStatePublicVariable)var;
            argval = pubvar.getVal();
            log("Variable "+pubvar.getName()+" is public and has value "+argval,4);
            secretdependent = false;
        }
        
    }
        
   if (secretdependent){
       VariableList thenvars =currstate.getCloneOfVariables();
       thenvars.setPublicVariable("pc", gotothen);
       VariableList elsevars =currstate.getCloneOfVariables();
       elsevars.setPublicVariable("pc", gotoelse);
    
       thenvars = processGuard(thenvars,varname,operator,expval,true);
       elsevars = processGuard(elsevars,varname,operator,expval,false);
       
    
       MarkovStatePrivateVariable hbefore =(MarkovStatePrivateVariable) currstate.getVariables().getVariable(varname);
       MarkovStatePrivateVariable hthen =(MarkovStatePrivateVariable) thenvars.getVariable(varname);
    
       BigDecimal thensize = new BigDecimal(hthen.getIntervalSize(),contexter.getContext());
       BigDecimal beforesize = new BigDecimal(hbefore.getIntervalSize(),contexter.getContext());
       BigDecimal thenprob = thensize.divide(beforesize,contexter.getContext());
       BigDecimal elseprob = contexter.getDecimalOne().subtract(thenprob,contexter.getContext());
    
       if (thenprob.compareTo(contexter.getDecimalZero())>0){
           MarkovChainState thenstate = this.getNextState(thenvars);  
           this.createTransition(currstate.getName(), thenstate.getName(), thenprob);
       }
    
       if (elseprob.compareTo(contexter.getDecimalZero())>0){
           MarkovChainState elsestate = this.getNextState(elsevars);
           this.createTransition(currstate.getName(), elsestate.getName(), elseprob);
       }   //
    
    
    

   }else{
       // the if does not depend on any private variables, and the value of the argument is known
       // since ifs don't have side effects, no variables except pc will be changed; we just need
       // to know if the next state will have the pc value of the then or of the else
       boolean guardistrue = false;
        switch (operator) {
            case "==":
                if (argval.compareTo(expval)==0) {
                    guardistrue = true;
                }
                break;
            case "!=":
                if (argval.compareTo(expval)!=0) {
                    guardistrue = true;
                }
                break;
            case ">=":
                if (argval.compareTo(expval)>=0) {
                    guardistrue = true;
                }
                break;
            case "<=":
                if (argval.compareTo(expval)<=0) {
                    guardistrue = true;
                }
                break;
            case "<":
                if (argval.compareTo(expval)<0) {
                    guardistrue = true;
                }
                break;
            case ">":
                if (argval.compareTo(expval)>0){
                    guardistrue = true;
                }
                break;
        }
       VariableList nextvars =currstate.getCloneOfVariables();
       if (guardistrue) {
           nextvars.setPublicVariable("pc", gotothen);
       }else{
           nextvars.setPublicVariable("pc", gotoelse);
       }

       MarkovChainState nextstate = this.getNextState(nextvars);  
       this.createTransition(currstate.getName(), nextstate.getName(), contexter.getDecimalOne());
   }
    
    
    
     
  }

    private String getFreshName() {
        String s = "s"+statecount;
        statecount++;
        return s;
    }

    private VariableList processGuard(VariableList vars, String varname, String op, BigInteger expval, boolean b) {
        String operator = op;
        if (logger.getVerbosity()>=4){
            log("Processing guard "+varname+operator+expval+" with variables "+vars.toString(),4);
        }

        // invert the operator if needed
        if (b==false){
            log("Inverting the operator as requested; original operator: "+operator,4);
            switch (operator) {
                case "==":
                    operator = "!=";
                    break;
                case "!=":
                    operator = "==";
                    break;
                case ">=":
                    operator = "<";
                    break;
                case "<=":
                    operator = ">";
                    break;
                case "<":
                    operator = ">=";
                    break;
                case ">":
                    operator = "<=";
                    break;
            }
            log ("New operator: "+operator,4);
        }
        
        // the guard is reduced to private variable -- operator -- constant; now we reduce the allowed intervals of the private variable accordingly
       MarkovStatePrivateVariable privvar = (MarkovStatePrivateVariable)vars.getVariable(varname);
       Interval interval = privvar.getInterval();
       interval.reduce(operator, expval);
       
      
       if (logger.getVerbosity()>=4){
           log("Setting intervals of variable "+privvar.getName()+" to "+privvar.getInterval(),4);
       }
        
        return vars;
    }

    

    private MarkovChainState getNextState(VariableList vars) {
        MarkovChainState newstate = stateset.getStateByVariableIdentifier(vars.getIdentifier());
    
        if (newstate == null){
            newstate = new MarkovChainState(this.getFreshName(),contexter.getDecimalZero());
            newstate.setVariables(vars);
            if (logger.getVerbosity()>=4){
                log("Built new state "+newstate.printNice()+"; adding it to the states to process",4);
            }
            toprocess.add(newstate);
        }else{
            if (logger.getVerbosity()>=4){
                log("Found state "+newstate.printNice()+" with identifier "+newstate.getVariables().getIdentifier(),4);
            }
        }
        return newstate;
    }

    private void createTransition(String s, String t, BigDecimal prob) {
        Transition trans = new Transition(s,t,prob);
        transset.add(trans);
        log("Created transition from state "+s+" to state "+t+" with probability "+prob,4);
    }

    

    private String getVariableName(String varname, VariableList variables) {
        String toreturn =varname;
        if (varname.contains("[")){
            String temp = varname.substring(varname.indexOf('['));
            while (temp.contains("[")){
                int closedpar = temp.indexOf(']');
                String index = temp.substring(1, closedpar);
                log("Evaluating index "+index,4);
                index = this.getVariableName(index, variables);
                BigInteger val = new BigInteger("0");
                try {
                    val = this.evaluateExpression(index, variables);
                } catch (ParseException e){
                    log(e.getMessage(),1);
                }
                toreturn = toreturn.replace("["+index+"]", "["+val.toString()+"]");
                temp = temp.substring(closedpar+1);
            }
                    
            
        }
        
        
        
        
        return toreturn;
    }

    private void processFreeLine(String aLine, BigInteger pc) {
       
    String varname;
        
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+QuailPatterns.pattern_free.pattern(),4);
    }
    
    Matcher matcher = QuailPatterns.pattern_free.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        this.toprocess.remove(currstate);
        parsingerror = true;
        return;
    }
    VariableList vars =currstate.getCloneOfVariables();
    stateset.add(currstate);
    toprocess.remove(currstate);
    
    log("Freed variable "+varname+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    vars.setPublicVariable("pc", newpc);
    vars.deleteVariableByName(varname);
    
    MarkovChainState newstate = getNextState(vars);
    
    createTransition(currstate.getName(),newstate.getName(),contexter.getDecimalOne());
    }

}

