/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author fbio
 */
class DiscriminationRelation {
    
    private final ArrayList<DiscriminationClass> relation;
    
    public DiscriminationRelation (){
        relation = new ArrayList<>();
    }
        
    public DiscriminationRelation (ArrayList<DiscriminationClass> listrel){
        relation = listrel;
    }
    
    public DiscriminationRelation (MarkovStateSet states){
        relation = new ArrayList<>(states.size());
        Iterator<MarkovChainState> statit = states.iterator();
        while (statit.hasNext()){
            DiscriminationClass aclass = new DiscriminationClass(1);
            MarkovChainState s =statit.next();
            aclass.add(s);
            aclass.setName("class_of_"+s.getName());
            relation.add(aclass);
        }
    }
    
    public DiscriminationRelation DiscriminateBy(String varname){
        DiscriminationRelation newrel = new DiscriminationRelation();
        
        for (int i=0; i<this.getNumberOfClasses(); i++){
            DiscriminationClass aclass = this.getClassByNumber(i);
            for (Object aclas : aclass) {
                // get a state
                MarkovChainState state = (MarkovChainState) aclas;
                MarkovStateVariable var = state.getVariable(varname);
                String classname;
                if (var==null){
                    classname = varname+"_unknown";
                }else{
                    classname = var.toString();
                }
                DiscriminationClass myclass = newrel.getClassByName(classname);
                if (myclass==null){
                    //case a: the class does not exist
                    myclass = new DiscriminationClass(classname,1);
                    myclass.add(state);
                    newrel.addClass(myclass);
                }else{
                    //case b: the class exists
                    myclass.add(state);
                }
            }
        }
        
        return newrel;
        
    }
    
    public DiscriminationRelation intersectBy(String varname){
        DiscriminationRelation newrel = new DiscriminationRelation();
        
        for (int i=0; i<this.getNumberOfClasses(); i++){
            DiscriminationClass aclass = this.getClassByNumber(i);
            for (int j=0; j<aclass.size(); j++){
                // get a state
                MarkovChainState state =(MarkovChainState) aclass.get(j);
                MarkovStateVariable var = state.getVariable(varname);
                
               
                String classname =aclass.getName()+",";
                if (var==null){
                   classname = classname+varname+"_unknown";
                }else{
                    classname += var.toString();
                }
                    
                DiscriminationClass myclass = newrel.getClassByName(classname);
                    
                    
                if (myclass==null){
                    //case a: the class does not exist
                    myclass = new DiscriminationClass(classname,1);
                    myclass.add(state);
                    newrel.addClass(myclass);
                }else{
                    //case b: the class exists
                    myclass.add(state);
                }
                
            }
        }
        
        return newrel;
    }
    
    public void addClass(DiscriminationClass theclass){
        relation.add(theclass);
    }
    
    public boolean addStateToClass(MarkovChainState state, String classname){
        DiscriminationClass theclass = this.getClassByName(classname);
        if (theclass==null){
            return false;
        }
        theclass.add(state);
        return true;
    }
    
    public int getNumberOfClasses(){
        return relation.size();
    }
    
    public DiscriminationClass getClassByNumber(int i){
        return relation.get(i);
    }
    
    public DiscriminationClass getClassOf(MarkovChainState state){
        DiscriminationClass toreturn = new DiscriminationClass();
        
        for (DiscriminationClass oneclass : relation) {
            for (Object oneclas : oneclass) {
                MarkovChainState s = (MarkovChainState) oneclas;
                if (s.getName().equals(state.getName())) {
                    toreturn = oneclass;
                }
            }
        }
        
        return toreturn;
    }
    
    public DiscriminationClass getClassByName(String name){
        DiscriminationClass toreturn = null;
        
        for (DiscriminationClass oneclass : relation) {
            if (oneclass.getName().equals(name)){
                toreturn = oneclass;
                break;
            }
        }
        return toreturn;
    }
    
    public void printNice(int level){
        log("Discrimination classes:",level);
        for (int i=0; i<this.getNumberOfClasses(); i++){
            
            DiscriminationClass thisclas = relation.get(i);
            String classtring = thisclas.getName()+": (";
            MarkovChainState s =(MarkovChainState) thisclas.get(0);
            classtring += s.getName();
            for (int j=1; j<thisclas.size(); j++){
                s =(MarkovChainState) thisclas.get(j);
                classtring += ",";
                classtring += s.getName();
            }
            classtring +=")";
            log(classtring,level);
        }
    }
    
    private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
}
