/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012-2014 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/

package quail;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.Set;

/**
 *
 * @author fbio
 */
public class IntervalAssignments extends HashMap<String, Interval>{

    IntervalAssignments(VariableList secrets) {
        for (MarkovStateVariable secretvar : secrets){
                Interval interv = ((MarkovStatePrivateVariable)secretvar).getInterval();
                this.put(secretvar.getName(), interv);
            }
    }
    
    IntervalAssignments(HashMap<String, Interval> map) {
        this.putAll(map);
    }

    public String printNice() {
        String toreturn = "";
        Set<String> keys = this.keySet();
        for (String key : keys){
            toreturn +=key+"="+this.get(key).toString();
        }
        return toreturn;
    }
    
    public boolean equals(IntervalAssignments other){
        for (String myvar:this.keySet()){
            if (!other.containsKey(myvar)){
                return false;
            }else if (!this.get(myvar).equals(other.get(myvar))){
                return false;
            }
        }
        
        return true;
    }
    
    public BigInteger getSize(){
        BigInteger toreturn = BigInteger.ZERO;
        for (Interval interv : this.values()){
            BigInteger thissize = interv.getSize();
            if (toreturn.compareTo(BigInteger.ZERO)==0){
                toreturn = thissize;
            }else{
                toreturn = toreturn.multiply(thissize);
            }
            
        }
        return toreturn;
    }
    
    boolean hasEmptyInterval() {
        for (Interval interv : this.values()){
            if (interv.isEmpty()){
                return true;
            }
        }
        return false;
    }
    
    public IntervalAssignments[] splitIntervals(IntervalAssignments newassignment) {
        IntervalAssignments[] toreturn = new IntervalAssignments[3];
        
        
        HashMap<String, Interval> mapintersection = new HashMap();
        HashMap<String, Interval> maprestofold = new HashMap();
        HashMap<String, Interval> maprestofnew = new HashMap();
        
        for (String varname : this.keySet()){
            Interval myinterv = this.get(varname);
            Interval hisinterv = newassignment.get(varname);
            Interval intersection = myinterv.intersect(hisinterv);
            
            mapintersection.put(varname, intersection);
            maprestofold.put(varname, myinterv.subtract(intersection));
            maprestofnew.put(varname,hisinterv.subtract(intersection));
        }
        
        // if the intersection has empty intervals, there are no shared assignments
        IntervalAssignments intersection = new IntervalAssignments(mapintersection);
        if (intersection.hasEmptyInterval()){
            toreturn[0] = intersection;
            toreturn[1] = this;
            toreturn[2] = newassignment;
            return toreturn;
        }
        
        
        toreturn[0] = new IntervalAssignments(mapintersection);
        toreturn[1] = new IntervalAssignments (maprestofold);
        toreturn[2] = new IntervalAssignments(maprestofnew);
        return toreturn;
    }
    
}
