/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.io.File;
import java.io.PrintWriter;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.logging.Level;
import org.apache.commons.math3.linear.Array2DRowFieldMatrix;
import org.apache.commons.math3.linear.ArrayFieldVector;
import org.apache.commons.math3.linear.FieldLUDecomposition;
import org.apache.commons.math3.linear.FieldVector;
import org.apache.commons.math3.util.BigReal;


/**
 *
 * @author Sorti
 */
public class MarkovChain implements Cloneable{
    protected MarkovStateSet states;
//    private HashSet transitions;
    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
    protected static final BigDecimal bogus_infinity = new BigDecimal("-2",contexter.getContext());


    public MarkovChain() {
       // this.transitions = transitions;
        this.states = null;    
    }
    
    public MarkovChain(MarkovStateSet states) {
       // this.transitions = transitions;
        this.states = states; 
        this.sanityCheck();
        
    }

    public MarkovChain(String[] sts, BigDecimal[] initprobs, BigDecimal[][] transs){
        this.buildStates(sts,initprobs,transs);
        
        
        this.sanityCheck();
    }

    MarkovChain(String[] states, double[] initprobs, double[][] trans) {
        int size = states.length;
        BigDecimal[] bigdecinitprobs = new BigDecimal[size];
        for (int i=0;i<size;i++){
            bigdecinitprobs[i] = new BigDecimal(initprobs[i],contexter.getContext());
        }
        
        BigDecimal[][] bigdectrans = new BigDecimal[size][size];
        for (int i=0; i<size; i++){
            for (int j=0; j<size; j++){
                bigdectrans[i][j] = new BigDecimal(trans[i][j],contexter.getContext());
            }
        }
        
        this.buildStates(states, bigdecinitprobs, bigdectrans);
        
        this.sanityCheck();
    }
    
    protected final void sanityCheck(){
        log("",4);
        log("==== Building a new Markov chain ====",4);
        log("Proposed state set: "+states,4);
        //log("Proposed transition set: "+transitions,4);
        /*//add all states to the IMC
        while(iter.hasNext()){
            Transition trans = (Transition)iter.next();
            log("Adding "+trans.toString()+" to the Markov chain");
            if (states.hasStateWithName(trans.getSource())==false) states.add(new MarkovChainState(trans.getSource(),0.0));
            if (states.hasStateWithName(trans.getDest())==false) states.add(new MarkovChainState(trans.getDest(),0.0));
           
        }*/
        for (MarkovChainState s : states) {
            HashSet<Transition> transitions = s.getTransitionsFrom();
            Iterator<Transition> transiter= transitions.iterator();
            boolean found = false;
            while(transiter.hasNext()){
                Transition trans = transiter.next();
                if (trans.getSource().equals(s.getName())) {
                    found = true;
                }
            }
            if (found==false){
                log("No outgoing transition found for state "+s+", assuming it's absorbing:",4);
                Transition trans = new Transition(s.getName(),s.getName(),contexter.getDecimalOne());
                log("Adding "+trans.toString()+" to the Markov chain",3);
                transitions.add(trans);
            } 
        }
        for (MarkovChainState s : states) {
            HashSet<Transition> trans = s.getTransitionsFrom();
            BigDecimal sum = contexter.getDecimalZero();
            Iterator<Transition> transiter;
            transiter = trans.iterator();
            while(transiter.hasNext()){
                Transition atrans =transiter.next();
                sum = sum.add(atrans.getValue());
            }
            
            if (sum.compareTo(contexter.getDecimalOne())==0){
                log("OK: The total outgoing probability from state "+s+" is 1",4);
            }else{
                log("WARNING: The total outgoing probability from state "+s+" is "+sum,3);
            }
        }
    }
    
    public MarkovStateSet getStates() {
        return states;
    }

    public void setStates(MarkovStateSet states) {
        this.states = states;
    }

    public HashSet<Transition> getTransitions() {
        HashSet<Transition> toreturn = new HashSet<>();
        for (MarkovChainState s : this.states) {
            HashSet<Transition> transitions = s.getTransitionsFrom();
            toreturn.addAll(transitions);
        }
        return toreturn;
    }

//    public void setTransitions(HashSet transitions) {
//        this.transitions = transitions;
//    }
    
    private static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }
    public HashSet<Transition> getTransitionsFrom(MarkovChainState s){
        return this.getTransitionsFrom(s.getName());
    }
    
    public HashSet<Transition> getTransitionsFrom(String s){
        if (states.hasStateWithName(s)==false){
            log("ERROR: requested transitions from nonexistent state "+s+"; will return empty set",1);
            return new HashSet<>();
        }
        HashSet<Transition> set = new HashSet<>();
        for (Transition trans : this.states.getStateByName(s).getTransitionsFrom()) {
            if (trans.getSource().equals(s)) {
                set.add(trans);
            }
        }
        if (set.isEmpty()){
            log("ERROR: state "+s+" has no outgoing transitions; this should not happen!",1);
        }
        return set;
    }
    
    public Transition getTransitionFromTo(MarkovChainState s, MarkovChainState t){
        return this.getTransitionFromTo(s.getName(), t.getName());
    }
    
    public Transition getTransitionFromTo(String s, String t){
        if (states.hasStateWithName(s)==false){
            log("Error: requested transitions from nonexistent state "+s+"; will return new transition with probability 0",1);
            return new Transition(s,t,contexter.getDecimalZero());
        }
        if (states.hasStateWithName(t)==false){
            log("Error: requested transitions to nonexistent state "+t+"; will return new transition with probability 0",1);
            return new Transition(s,t,contexter.getDecimalZero());
        }
        return this.states.getStateByName(s).getTransitionToState(states.getStateByName(t));
        
    }
    
    public BigDecimal getTransitionProbabilityFromTo(MarkovChainState s, MarkovChainState t){
        return this.getTransitionProbabilityFromTo(s.getName(), t.getName());
    }
    
    public BigDecimal getTransitionProbabilityFromTo(String s, String t){
        Transition trans = this.getTransitionFromTo(s, t);
        return trans.getValue();
    }
    
    public void removeTransitionFromTo(String s, String t){
        this.removeTransitionFromTo(states.getStateByName(s), states.getStateByName(t));
    }
    
    private void removeTransitionFromTo(MarkovChainState source, MarkovChainState dest) {
        Transition trans = this.getTransitionFromTo(source, dest);
        if (logger.getVerbosity()>=4){
            log("Removing transition from state "+source.getName()+" to state "+dest.getName(),4);
        }
        source.removeTransition(trans);
        source.removeSuccessor(dest);
        dest.removePredecessor(source);
    }
    
    private void removeTransition(Transition trans) {
        MarkovChainState source = states.getStateByName(trans.getSource());
        MarkovChainState dest = states.getStateByName(trans.getDest());
        if (logger.getVerbosity()>=4){
            log("Removing transition from state "+source+" to state "+dest,4);
        }
        source.removeTransition(trans);
        source.removeSuccessor(dest);
        dest.removePredecessor(source);
    }
    
    private void addTransition(Transition trans) {
        if (trans.getValue().compareTo(contexter.getDecimalZero())>0){
            MarkovChainState source = states.getStateByName(trans.getSource());
            MarkovChainState dest = states.getStateByName(trans.getDest());
            if (logger.getVerbosity()>=4){
                log("Adding transition from state "+source+" to state "+dest+" with probability "+trans.getValue(),4);
            }
            source.addTransition(trans);
            source.addSuccessor(dest);
            dest.addPredecessor(source);
        }
    }
/*    public Boolean satisfies(IntervalMarkovChain imc){
        log("Checking if the Markov chain satisfies the IMC");
        HashSet imcstates = imc.getStates(); */
        /*if (imcstates.equals(this.states)==false){
            log("The Markov chain does not satisfy the IMC because the state sets are different");
            return false;
        } TODO: reimplement*/
        
        /*I just check that the intervals are satisfied; this is sufficient under two assumptions:
         * 1- the Markov chain is a Markov chain (all outgoing probabilities sum up to 1)
         * 2- the IMC is satisfiable
         */
  /*      Iterator stateiter = states.iterator();
        while (stateiter.hasNext()){
            MarkovChainState s =(MarkovChainState) stateiter.next();
            HashSet transfroms = imc.getTransitionIntervalsFrom(s.getName());
            Iterator transfromsiter = transfroms.iterator();
            while (transfromsiter.hasNext()){
                IntervalTransition interval =(IntervalTransition) transfromsiter.next();
                String t =(String) interval.getDest();
                double fromstot = this.getTransitionProbabilityFromTo(s.getName(),t);
                if (interval.getBottom()>fromstot || interval.getTop()<fromstot){
                    log("The Markov chain does not satisfy the IMC because the transition from "+s+" to "+t+" has probability "+fromstot+" but the allowed interval is ["+interval.getBottom()+","+interval.getTop()+"]");
                    return false;
                }
            }
        }
        log("The Markov chain satifies the IMC");
        return true;
    }*/
    /*
     * codice generico per risolvere una matrice:
     * double[][] coefficients = { { 3.0, 20.0, 89.0 },
                            { 4.0, 40.0, 298.0 },
                            { 7.0, 21.0, 0.42 } };
       double[] values = { 1324, 2999, 2039 };
       RealMatrix matrix = new RealMatrixImpl( );
       matrix.setData( coefficients );
 
       double[] answers = matrix.solve( values );
       System.out.println( "Answers: " + ArrayUtils.toString( answers ) );
     * 
     */
    
    public MarkovChainState getStateByName(String name){
        return states.getStateByName(name);
    }
    
    public MarkovChain computeQuotient(DiscriminationRelation rel){
        int qsize = rel.getNumberOfClasses();
        log("Building a quotient with "+qsize+" classes",2);
        String[] qstnames = new String[qsize];
        BigDecimal[] qinitprob = new BigDecimal[qsize];
        for (int i=0; i<qsize; i++){
            qinitprob[i]=contexter.getDecimalZero();
        }
        BigDecimal[][] qtrans = new BigDecimal[qsize][qsize];
        for (int i=0; i<qsize; i++){
            for (int j=0; j<qsize; j++){
                qtrans[i][j]=contexter.getDecimalZero();
            }
        }
        
        
        
        for (int classnumber=0; classnumber<qsize; classnumber++){
            DiscriminationClass sclas;
            sclas = rel.getClassByNumber(classnumber);
            //get a state
            MarkovChainState s;
            s = (MarkovChainState)sclas.get(0);
            
            //define class name
            String sclasname = s.getName();
            if (sclas.size()>1){
                for (int i=1; i<sclas.size(); i++) {
                    sclasname = sclasname + "," + ((MarkovChainState)sclas.get(i)).getName();
                }
            }
            qstnames[classnumber] = sclasname;
            log("The name of class "+classnumber+" is "+sclasname,4);
            
            //compute class initial probability
            BigDecimal sclasinitprob = contexter.getDecimalZero();
            for (Object scla : sclas) {
                sclasinitprob = sclasinitprob.add(((MarkovChainState) scla).getInitialProbability());
            }
            qinitprob[classnumber] = sclasinitprob;
            log("The initial probability of class "+classnumber+" is "+sclasinitprob,4);
            
            //compute transition probabilities for this class
            /*
             * for each class [s] the transition probability to each class [t] is the sum of the transition probabilities
             * from each state of [s] to each state of [t], all divided by the size of [s]
             */
            int sclassize =sclas.size();
            for (int i=0; i<qsize; i++){
                //i is the number of the class [t]
                DiscriminationClass tclas = rel.getClassByNumber(i);
                BigDecimal probstot = contexter.getDecimalZero();
                for (Object scla : sclas) {
                    for (Object tcla : tclas) {
                        MarkovChainState mys = (MarkovChainState) scla;
                        MarkovChainState myt = (MarkovChainState) tcla;
                        probstot = probstot.add(this.getTransitionProbabilityFromTo(mys.getName(), myt.getName()));
                    }
                }
                // divide the probability by the size of [s]
                probstot = probstot.divide(new BigDecimal(sclassize,contexter.getContext()),contexter.getContext());
                
                // write it in the probability transition matrix of the quotient
                qtrans[classnumber][i] = probstot;
            }
        }
        
        
        MarkovChain quot = new MarkovChain(qstnames, qinitprob, qtrans);
        
       return quot;
        
    }
    
    
    public BigDecimal calculateEntropy(){
      log("Calculating entropy of the Markov chain...",3);
        
      for (MarkovChainState s:states){
          s.setReward(this.getLocalEntropyOfState(s));
      }  
      
      return this.calculateExpectedReward();
    }
    
    public BigDecimal calculateEntropyRate(){
      log("Calculating entropy rate of the Markov chain...",3);
        
      for (MarkovChainState s:states){
          s.setReward(this.getLocalEntropyOfState(s));
      }  
      
      return this.calculateExpectedRewardRate();
    }
           

    public void printNice(int level) {
        //print the state space
        String s="States: (";
        Iterator<MarkovChainState> stateiterator = states.iterator();
        if (stateiterator.hasNext()){
            MarkovChainState state = stateiterator.next();
            s += state.printNice();
            
            
        } 
            
        while (stateiterator.hasNext()){
            MarkovChainState state = stateiterator.next();
            s = s+" , "+state.printNice();
                       
        }
        s += ");";
        log(s,level);
        
        // print the transitions for each state
        log("Transitions:",level);
        stateiterator = states.iterator();
        while (stateiterator.hasNext()){
            String state = stateiterator.next().getName();
            HashSet<Transition> transofs = this.getTransitionsFrom(state);
            Iterator<Transition> transiter = transofs.iterator();
            while (transiter.hasNext()){
                Transition trans =transiter.next();
                if (trans.getValue().compareTo(contexter.getDecimalZero())>0) {
                    log("  "+trans,level);
                }
            }
        }
        log("",level);
    }
    
    private HashSet<HashSet<String>>[] calculateSCCs(){
        Tarjan tarj = new Tarjan();
        tarj.initialize(this);
        HashSet<HashSet<String>> SCCs = tarj.getSCC();
        log("The SCCs are "+SCCs,4);
        //SCCs that have probability 0 of transitioning outside themselves are recurrent, others are transient. States in a recurrent SCC have infinite expected occupation time.
        Iterator<HashSet<String>> scciter = SCCs.iterator();
        HashSet<HashSet<String>> transientSCCs = new HashSet<>();
        HashSet<HashSet<String>> recurrentSCCs = new HashSet<>();
        while (scciter.hasNext()){
            //SCC is a HastSet of HastSet of String
            HashSet<String> myscc = scciter.next();
            if (this.isRecurrentSCC(myscc)){
                recurrentSCCs.add(myscc);
            } 
            else{
                transientSCCs.add(myscc);
            }
                
        }
        log("Recurrent SCCs: "+recurrentSCCs,4);
        log("Transient SCCs: "+transientSCCs,4);
        HashSet<HashSet<String>>[] toreturn;
        toreturn = new HashSet[2];
        toreturn[0] = recurrentSCCs;
        toreturn[1] = transientSCCs;
        return toreturn;
        
    }
    
    private void calculateXis(){
        log("Calculating the xi of each state; this should be called only once per Markov chain",3);
        HashSet[] SCCs = this.calculateSCCs();
        HashSet<HashSet<String>> recurrentSCCs = SCCs[0];
        HashSet<HashSet<String>> transientSCCs = SCCs[1];
        
        // setting xirate to infinite for all recurrent states
        Iterator<HashSet<String>> recscciter = recurrentSCCs.iterator();
        while (recscciter.hasNext()){
            HashSet<String> recset =recscciter.next(); 
            Iterator<String> recstateiter = recset.iterator();
            while (recstateiter.hasNext()){
                String recstate =recstateiter.next();
                MarkovChainState s = states.getStateByName(recstate);
                s.setXi(bogus_infinity);
            }
        }
        // if there are no transient states, nothing else to do
        if (transientSCCs.isEmpty()) {
            return;
        }
        //preparing the system of linear equations to calculate the xirate of all transient states
        ArrayList<String> transientstates = new ArrayList<>();
        Iterator<HashSet<String>> transscciter = transientSCCs.iterator();
        while (transscciter.hasNext()){
            HashSet<String> transset =transscciter.next(); 
            Iterator<String> transstateiter = transset.iterator();
            while (transstateiter.hasNext()){
                String transstate =transstateiter.next();
                transientstates.add(transstate);
            }
        }
        
        int matrixsize = transientstates.size();
        log("The "+matrixsize+" transient states are "+transientstates,4);
        
        /* the equation system looks like:
         * (P_ss - 1)*xi_s + P_ts*xi_t + P_us*xi_u = -pi^0_s
         * P_st*xi_s + (P_tt-1)*xi_t + P_ut*xi_u = -pi^0_t
         * P_su*xi_s + P_tu*xi_t + (P_uu-1)*xi_u = -pi^0_u
         * solve it to obtain all xis
         */
        Array2DRowFieldMatrix<BigReal> matrix;
        BigReal[][] forthematrix = new BigReal[matrixsize][matrixsize];
        for (int i=0; i<matrixsize; i++){
            String statefrom = transientstates.get(i);
            for (int j=0; j<matrixsize; j++){
                String stateto = transientstates.get(j);
                BigReal transprob = new BigReal(this.getTransitionProbabilityFromTo(statefrom,stateto));
                if (i==j) {
                    forthematrix[i][j]= transprob.subtract(BigReal.ONE);
                }
                else {
                    forthematrix[i][j]= transprob;
                }
                if (transprob.compareTo(BigReal.ZERO)!=0){
                    log("Transition probability "+transprob+" from "+statefrom+" to "+stateto+" added to the matrix",4);
                }
            }
            
            //String printablerow = Double.toString(row[0]);
            //for (int u=1;u<matrixsize;u++) printablerow = printablerow+","+Double.toString(row[u]);
            //log("Adding line number "+i+" of the matrix: "+printablerow,4);
        }
        matrix = new Array2DRowFieldMatrix<>(forthematrix);
        
        //log("The transition matrix for the transient states is "+matrix.toString(),4);
        //preparing the coefficient vector (with the initial probabilities)
        
        BigReal[] forthevector = new BigReal[matrixsize];
        
        for (int i=0; i<matrixsize; i++){
            MarkovChainState state =states.getStateByName(transientstates.get(i));
            forthevector[i]= new BigReal(contexter.getDecimalZero().subtract(state.getInitialProbability()));
        }
        ArrayFieldVector<BigReal> coeff;
        coeff = new ArrayFieldVector<>(forthevector);
        //String printablecoeff = coeff.getEntry(0).toString();
        //  for (int u=1;u<matrixsize;u++) {
        //      printablecoeff = printablecoeff+","+coeff.getEntry(u).toString();
        //  }
        //log("The coefficients are "+printablecoeff,4);
        //finally, solving the equation system to get the xirate of each transient state
        FieldLUDecomposition<BigReal> solver = new FieldLUDecomposition<>(matrix);
        FieldVector<BigReal> solutions = solver.getSolver().solve(coeff);
        //String printablesol = solutions.getEntry(0).toString();
        //  for (int u=1;u<matrixsize;u++) {
        //      printablesol = printablesol+","+solutions.getEntry(u).toString();
        //  }
        //log("The solutions are "+printablesol);
        
        for (int i=0;i<matrixsize;i++){
            String statename = transientstates.get(i);
            // I hope casting does not lose precision; BigDecimal is just a wrapper of BigDecimal AFAIK
            BigDecimal xi =solutions.getEntry(i).bigDecimalValue();
            xi.round(contexter.getContext());
            MarkovChainState state = states.getStateByName(statename);
            state.setXi(xi);
            log("State "+state+" has xi "+xi,4);
        }
        
    }
    
    private void calculateXiRates(){
        log("Calculating the xi rate of each state; this should be called only once per Markov chain",3);
        HashSet[] SCCs = this.calculateSCCs();
        HashSet<HashSet<String>> recurrentSCCs = SCCs[0];
        HashSet<HashSet<String>> transientSCCs = SCCs[1];
        MarkovChain clonedchain = null;
        try {
            clonedchain = this.clone();
        } catch (CloneNotSupportedException ex) {
            java.util.logging.Logger.getLogger(MarkovChain.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);        
        }
        
        if (clonedchain == null){
            log("ERROR: unable to clone Markov chain",1);
            System.exit(1);
        }
        // setting xirate rate to zero for all transient states
        for (HashSet<String> statenames : transientSCCs) { 
            MarkovStateSet settohide = new MarkovStateSet();
            for (String statename: statenames){
                MarkovChainState statetohide = this.getStateByName(statename);
                settohide.add(statetohide);
                statetohide.setXiRate(contexter.getDecimalZero());
                
            }
            clonedchain.hideStates(settohide);
        }
        for (HashSet<String> recstatesnames : recurrentSCCs) {
            log("Computing xi rates for class "+recstatesnames.toString(),4);
            MarkovStateSet recset = new MarkovStateSet(); 
            for (String statename: recstatesnames){
                recset.add(this.getStateByName(statename));
            }
            
            if (recset.size()==1){
                
                MarkovChainState recstate =recset.iterator().next();
                
                BigDecimal initprob = recstate.getInitialProbability();
                recstate.setXiRate(initprob);
                log("Class has a single state; its xi rate equals its initial probability of "+initprob.toEngineeringString(),4);
                continue;
            }
            
            //preparing the system of linear equations to calculate the xirate rate of the recurrent states in this SCC
            ArrayList<String> recurrentstates = new ArrayList<>();
            BigDecimal totclassprob = contexter.getDecimalZero();
            for (MarkovChainState recstate:recset){
                recurrentstates.add(recstate.getName());
                MarkovChainState clonedstate = clonedchain.getStateByName(recstate.getName());
                totclassprob = totclassprob.add(clonedstate.getInitialProbability(), contexter.getContext());
            }

            int matrixsize = recurrentstates.size();
            log("The class has "+matrixsize+" recurrent states "+recurrentstates+" and probability "+totclassprob.toEngineeringString(),4);

            /* 
            * the equation system looks like:
            * line 0:
            * 0 + \xi_1(P_10-P_00+1)+\xi_2(P_20-P_00+1)+...+\xi_n(P_n0-P_00+1) = -c(P_00-1)
            * line i for i>0:
            * 0 + \xi_1(P_1i-P_i0)+\xi_2(P_2i-Pi0)+...+\xi_i(P_ii-P_i0-1)+...+\xi_n(P_ni-Pi0) = -c P_i0
            * solve it to obtain all xi rates
            */
            Array2DRowFieldMatrix<BigReal> matrix;
            BigReal[][] forthematrix = new BigReal[matrixsize][matrixsize];
            for (int i=0; i<matrixsize; i++){
                String statei = recurrentstates.get(i);
                forthematrix[i][0] = BigReal.ZERO;
                BigReal prob_i_to_0 = new BigReal(this.getTransitionProbabilityFromTo(statei,recurrentstates.get(0)));
                for (int j=1; j<matrixsize; j++){
                    String statej = recurrentstates.get(j);
                    BigReal prob_j_to_i = new BigReal(this.getTransitionProbabilityFromTo(statej,statei));
                    if (i==j) {
                        forthematrix[i][j]= prob_j_to_i.subtract(prob_i_to_0).subtract(BigReal.ONE);
                    }

                    else {
                        forthematrix[i][j]= prob_j_to_i.subtract(prob_i_to_0);
                    }
                    if (i==0){
                        forthematrix[i][j] = forthematrix[i][j].add(BigReal.ONE);
                    }
                    //if (transprob.compareTo(BigReal.ZERO)!=0){
                    //    log("Transition probability "+transprob.bigDecimalValue().toEngineeringString()+" from "+statei+" to "+statej+" added to the matrix",4);
                    //}
                }
            }

                /*String printablerow = forthematrix[i][0];
                for (int u=1;u<matrixsize;u++) printablerow = printablerow+","+Double.toString(row[u]);
                log("Adding line number "+i+" of the matrix: "+printablerow,4);*/
            
            matrix = new Array2DRowFieldMatrix<>(forthematrix);

            log("The matrix is:",4);
            for (int i=0; i<matrixsize; i++){
                String line = forthematrix[i][0].bigDecimalValue().toEngineeringString();
                for (int j=1; j<matrixsize; j++){
                    line = line+" , "+forthematrix[i][j].bigDecimalValue().toEngineeringString();    
                }
                log(line,4);
            }      
            
            
            //log("The transition matrix for the recurrent class is "+matrix.toString(),4);
            /* preparing the coefficient vector. Let c be the total class probability. 
             * The coefficients are:
             * -c (P_00 -1)
             * -c (P_10)
             * ...
             * -c (P_n0)
             */
            BigReal[] forthevector = new BigReal[matrixsize];
            
            BigReal realclassprob = new BigReal(totclassprob);
            BigReal prob_0_to_0 = new BigReal(this.getTransitionProbabilityFromTo(recurrentstates.get(0),recurrentstates.get(0)));
            forthevector[0] = BigReal.ZERO.subtract(realclassprob.multiply(prob_0_to_0.subtract(BigReal.ONE)));
            for (int i=1; i<matrixsize; i++){
                //MarkovChainState state =(MarkovChainState) states.getStateByName(recurrentstates.get(i));
                //forthevector[i]= new BigReal(contexter.getDecimalZero().subtract(state.getInitialProbability()));
                BigReal prob_i_to_0 = new BigReal(this.getTransitionProbabilityFromTo(recurrentstates.get(i),recurrentstates.get(0)));
                forthevector[i] = BigReal.ZERO.subtract(realclassprob.multiply(prob_i_to_0));
                
            }
            ArrayFieldVector<BigReal> coeff;
            coeff = new ArrayFieldVector<>(forthevector);
            String printablecoeff = coeff.getEntry(0).bigDecimalValue().toEngineeringString();
              for (int u=1;u<matrixsize;u++) {
                  printablecoeff = printablecoeff+" , "+coeff.getEntry(u).bigDecimalValue().toEngineeringString();
              }
            log("The coefficients are "+printablecoeff,4);
            
            //finally, solving the equation system to get the xirate of each transient state
            FieldLUDecomposition<BigReal> solver = new FieldLUDecomposition<>(matrix);
            FieldVector<BigReal> solutions = solver.getSolver().solve(coeff);
            //String printablesol = solutions.getEntry(0).toString();
            //  for (int u=1;u<matrixsize;u++) {
            //      printablesol = printablesol+","+solutions.getEntry(u).toString();
            //  }
            //log("The solutions are "+printablesol);

            for (int i=0;i<matrixsize;i++){
                String statename = recurrentstates.get(i);
                // I hope casting does not lose precision; BigReal is just a wrapper of BigDecimal AFAIK
                BigDecimal xirate =solutions.getEntry(i).bigDecimalValue();
                xirate.round(contexter.getContext());
                MarkovChainState state = states.getStateByName(statename);
                state.setXiRate(xirate);
                log("State "+state+" has xi rate "+xirate,4);
            }
        }
        
        
        
    }
    
    
    private void calculateLocalEntropies(){
        log("Calculating local entropy of each state; this should be called only once per Markov chain",3);
        for (MarkovChainState state : states) {
            String s = state.toString();
            log("Calculating local entropy of state "+s+"...",4);
            HashSet<Transition> successors = this.getTransitionsFrom(s);
            BigDecimal localentropy = contexter.getDecimalZero();
            Iterator<Transition> succiter = successors.iterator();
            
            while (succiter.hasNext()){
                String t =succiter.next().getDest();
                BigDecimal prob = this.getTransitionProbabilityFromTo(s, t);
                if (prob.compareTo(contexter.getDecimalZero())>0 && prob.compareTo(contexter.getDecimalOne())<0){
                    // add -(prob* log_2 (prob)) to the local expectedrewardrate
                    log("The transition probability is "+prob.toPlainString(),4);
                    BigDecimal log2 = this.log2(prob);
                    log("Its logarithm in base 2 is "+log2.toPlainString(),4);
                    BigDecimal minuslog2 = contexter.getDecimalZero().subtract(log2);
                    BigDecimal toadd = prob.multiply(minuslog2);
                    localentropy = localentropy.add(toadd);
                    log("The transition from "+s+" to "+t+" contributes "+toadd+" bits of local entropy",4);
                }
            }
            log("State "+s+" has local entropy "+localentropy.toPlainString(),4);
            state.setLocalEntropy(localentropy);
        }
        
    }
    
    private void calculateRecurrencies(){
        for (MarkovChainState state : states) {
            if (this.getXiOfState(state.getName())==bogus_infinity){
                state.setRecurrent(true);
            }else {
                state.setRecurrent(false);
            }
        }
    }

    
    
    public BigDecimal getXiOfState(MarkovChainState s){
        if (s.getXi().compareTo(new BigDecimal("-1",contexter.getContext()))==0) {
            this.calculateXis();
        }
        return s.getXi();
    }
    
    public BigDecimal getXiRateOfState(String name){
        MarkovChainState s = states.getStateByName(name);
        return this.getXiRateOfState(s);
    }
    
    public BigDecimal getXiRateOfState(MarkovChainState s){
        if (s.getXiRate().compareTo(new BigDecimal("-1",contexter.getContext()))==0) {
            this.calculateXiRates();
        }
        return s.getXiRate();
    }
    
    public BigDecimal getXiOfState(String name){
        MarkovChainState s = states.getStateByName(name);
        return this.getXiOfState(s);
    }
    
    public BigDecimal getLocalEntropyOfState(MarkovChainState s){
        if (s.getLocalEntropy().compareTo(new BigDecimal("-1",contexter.getContext()))==0) {
            this.calculateLocalEntropies();
        }
        return s.getLocalEntropy();
    }
    
    public BigDecimal getLocalEntropyOfState(String name){
        MarkovChainState s = states.getStateByName(name);
        return this.getLocalEntropyOfState(s);
    }
    
    public boolean isRecurrentState (MarkovChainState s){
        if (s.getXi().compareTo(new BigDecimal("-1",contexter.getContext()))==0) {
            this.calculateRecurrencies();
        }
        return s.isRecurrent();
    }
    
    public boolean isRecurrentState (String name){
        MarkovChainState s = states.getStateByName(name);
        return this.isRecurrentState(s);
    }
    
    public boolean isDeterministicState (MarkovChainState s){
        return this.getLocalEntropyOfState(s).compareTo(contexter.getDecimalOne())==0;
    }
    
    public boolean isDeterministicState (String name){
        MarkovChainState s = states.getStateByName(name);
        return this.isDeterministicState(s);
    }
    
    private boolean isRecurrentSCC(HashSet<String> myscc) {
        Iterator<String> scciter = myscc.iterator();
        while (scciter.hasNext()){
            String s =scciter.next();
            Iterator<MarkovChainState> statesiter = states.iterator();
            while (statesiter.hasNext()){
                String t =statesiter.next().getName();
                if (myscc.contains(t)== false){
                    if (this.getTransitionProbabilityFromTo(s, t).compareTo(contexter.getDecimalZero())>0){
                        return false;
                    }
                }
                    
                        
            }
            
        }
        return true;
    }

    void hideStates(MarkovStateSet tohide) {
        Iterator<MarkovChainState> tohideiter = tohide.iterator();
        boolean divergenceisreachable = false;
        MarkovChainState divergstate = new MarkovChainState("divergence",contexter.getDecimalZero());
        MarkovStatePublicVariable div_observ = new MarkovStatePublicVariable("observe_divergence",BigInteger.ZERO,1);
        div_observ.setObservable(true);
        divergstate.addVariable(div_observ);

        MarkovStatePrivateVariable div_secret = new MarkovStatePrivateVariable("secret_divergence","[0,1]",1);
        div_secret.setSecret(true);
        divergstate.addVariable(div_secret);

        states.add(divergstate);
        Transition divloop = new Transition (divergstate.getName(),divergstate.getName(),contexter.getDecimalOne());
        this.addTransition(divloop);
        
        while (tohideiter.hasNext()){
            // get the state to hide
            MarkovChainState statetohide =tohideiter.next();
            log("",3);
            log("Hiding state "+statetohide.getName(),4);
            
            // get the looping probability
            BigDecimal loopprob = this.getTransitionProbabilityFromTo(statetohide.getName(),statetohide.getName());
            
            // build the set of predecessors
            MarkovStateSet predecessors = statetohide.getPredecessors();
//            Iterator stateiter = states.iterator();
//            while (stateiter.hasNext()) {
//                MarkovChainState pred = (MarkovChainState)stateiter.next();
//                if (this.getTransitionProbabilityFromTo(pred.getName(), statetohide.getName()).compareTo(contexter.getDecimalZero())>0){
//                    if (pred != statetohide){
//                        log("Found the predecessor state "+pred.getName()+"; the transition probability from "+pred.getName()+" to "+statetohide.getName()+" is "+this.getTransitionProbabilityFromTo(pred.getName(), statetohide.getName()),4);
//                        predecessors.add(pred);
//                    }
//                }
//            }
            
            if (predecessors.isEmpty()){
                // this assumes that the initial state cannot be hidden...
                log("Hiding state "+statetohide.getName()+" with no predecessors",4);
            }else{
                if (logger.getVerbosity()>=4){
                    log("Its predecessors are "+predecessors.toString(),4);
                }
            }
            
            
            if (loopprob.compareTo(contexter.getDecimalOne())==0){
                log("It is an absorbing state; redirecting all predecessors to divergence",3);
                divergenceisreachable = true;
                MarkovStateSet predtoprocess = new MarkovStateSet();
                predtoprocess.addAll(predecessors);
                for (MarkovChainState pred : predtoprocess) {
                    Transition transtodiv = this.getTransitionFromTo(pred.getName(), divergstate.getName());
                    Transition transtostate = this.getTransitionFromTo(pred.getName(), statetohide.getName());
                    BigDecimal transprob = transtostate.getValue();
                    if (transtodiv.getValue().compareTo(contexter.getDecimalZero())==0){
                        transtodiv.setValue(transprob);
                        this.addTransition(transtodiv);
                        if (logger.getVerbosity()>=4){
                            log("Created transition from state "+pred.getName()+" to divergence with probability "+transprob,4);
                        }
                    }else{
                        transtodiv.setValue(transprob.add(transtodiv.getValue()));
                        if (logger.getVerbosity()>=4){
                            log("Increased divergence probability for state "+pred.getName()+" to "+transtodiv.getValue(),4);
                        }
                    }
                    this.removeTransition(transtostate);
                }
            }else{
            	log("Loop probability: "+loopprob.toPlainString()+" with precision "+loopprob.precision()+" and scale "+loopprob.scale(),4);
                
            	BigDecimal normprob = contexter.getDecimalOne().subtract(loopprob);
                log("Normalized probability: "+normprob.toPlainString()+" with precision "+normprob.precision()+" and scale "+normprob.scale(),4);
               
                // build the set of successors
                MarkovStateSet successors = statetohide.getSuccessors();
//                        new MarkovStateSet();
//                stateiter = states.iterator();
//                while (stateiter.hasNext()) {
//                    MarkovChainState succ = (MarkovChainState)stateiter.next();
//                    if (this.getTransitionProbabilityFromTo(statetohide.getName(), succ.getName()).compareTo(contexter.getDecimalZero())>0){
//                        if (succ != statetohide){
//                            log("Found the successor state "+succ.getName()+"; the transition probability from "+statetohide.getName()+" to "+succ.getName()+" is "+this.getTransitionProbabilityFromTo(statetohide.getName(), succ.getName()),4);
//                            successors.add(succ);
//                        }
//                        
//                    }
//                }
                
                if (successors.isEmpty()){
                    log("ERROR: trying to hide state "+statetohide.getName()+" but it has no successors",1);
                }else{
                    if (logger.getVerbosity()>=4){
                        log("Its successors are "+successors.toString(),4);
                    }
                }
                
                for (MarkovChainState succ:successors){
                    Transition transfromstate = this.getTransitionFromTo(statetohide.getName(), succ.getName());
                    BigDecimal transfromstateprob = transfromstate.getValue();
                    BigDecimal eventualprob = transfromstateprob.divide(normprob,contexter.getContext());
                    BigDecimal newinitprob = succ.getInitialProbability().add(statetohide.getInitialProbability().multiply(eventualprob, contexter.getContext()), contexter.getContext());
                    log("Updating initial probability of state "+succ.getName()+" from "+succ.getInitialProbability().toEngineeringString()+" to "+newinitprob.toEngineeringString(),4);
                    succ.setInitialProbability(newinitprob);
                }
                
                
                MarkovStateSet predtoprocess = new MarkovStateSet();
                predtoprocess.addAll(predecessors);
                Iterator<MarkovChainState> prediter = predtoprocess.iterator();
                while (prediter.hasNext()){
                    MarkovChainState pred;
                    pred = prediter.next();
                    Transition transtostate = this.getTransitionFromTo(pred.getName(), statetohide.getName());
                    BigDecimal transtostateprob = transtostate.getValue();
                        for (MarkovChainState succ : successors) {
                            Transition transfromstate = this.getTransitionFromTo(statetohide.getName(), succ.getName());
                            BigDecimal transfromstateprob = transfromstate.getValue();
                            BigDecimal eventualprob = transfromstateprob.divide(normprob,contexter.getContext());
                            
                            Transition longtrans = this.getTransitionFromTo(pred.getName(), succ.getName());
                            BigDecimal longtransprob = longtrans.getValue().add(transtostateprob.multiply(eventualprob,contexter.getContext()));
                            if (longtrans.getValue().compareTo(contexter.getDecimalZero())==0){
                                longtrans.setValue(longtransprob);
                                this.addTransition(longtrans);
                            }else{
                                if (logger.getVerbosity()>=4){
                                    log("Increasing transition probability from state "+pred.getName()+" to state "+succ.getName()+" from "+longtrans.getValue()+" to "+longtransprob,4);
                                }
                                longtrans.setValue(longtransprob);
                            }
                        }
                    
                    // if this is the last predecessor, remove the transitions from the state to hide to its successors
                    if (prediter.hasNext()==false){
                        MarkovStateSet toremove = new MarkovStateSet();
                        toremove.addAll(successors);
                        for (MarkovChainState succ : toremove) {
                            this.removeTransitionFromTo(statetohide, succ);
                        }
                    }
                    
                    this.removeTransition(transtostate);
                }
                
            }
            if (loopprob.compareTo(contexter.getDecimalZero())>0){
                Transition loop = this.getTransitionFromTo(statetohide.getName(), statetohide.getName());
                this.removeTransition(loop);
            }
            //it could be reasonable to implement some sanity checks here, like that no transition from or to the state to hide is left
            
            states.remove(statetohide);
            log("Hidden state "+statetohide.getName(),3);
            
        }
        
        
        
        if (divergenceisreachable == false){
            states.remove(divergstate);
            //transitions.remove(divloop);
            log("Divergence state is unreachable; removing it",3);
        }
        
        this.resetCalculations();
    }

    void resetCalculations() {
        log("Resetting residence time and local entropy of each state...",3);
        for (MarkovChainState s : states) {
            s.setXi(new BigDecimal("-1",contexter.getContext()));
            s.setLocalEntropy(new BigDecimal("-1",contexter.getContext()));
        }
        log("Done.",3);
    }

    ArrayList<String> getVariablesStartingWith(String start) {
        // never used, I leave it here for legacy purposes
        Iterator<MarkovChainState> statiter = states.iterator();
        ArrayList<String> toreturn = new ArrayList<>();
        while (statiter.hasNext()){
            MarkovChainState s = statiter.next();
            VariableList vars = s.getVariables();
            for (MarkovStateVariable var : vars) {
                String varname = var.getName();
                if (varname.startsWith(start)){
                    if (toreturn.contains(varname)==false){
                        toreturn.add(var.getName());
                    }
                    
                }
            }
        }
        
        return toreturn;
    }

    ArrayList<String> getObservableVariables() {
        Iterator<MarkovChainState> statiter = states.iterator();
        ArrayList<String> toreturn = new ArrayList<>();
        while (statiter.hasNext()){
            MarkovChainState s = statiter.next();
            VariableList vars = s.getVariables();
            for (MarkovStateVariable var : vars) {
                if (var.isPrivate()==false){
                    MarkovStatePublicVariable pubvar =(MarkovStatePublicVariable)var;
                    //log("Found public variable "+pubvar.getName(),4);
                    if (pubvar.isObservable()){
                        //log("Found observable variable "+pubvar.getName(),4);
                        if (toreturn.contains(pubvar.getName())==false){
                            toreturn.add(var.getName());
                        }
                    }
                }
            }
        }
        
        return toreturn;
    }
    
    ArrayList<String> getSecretVariables() {
        Iterator<MarkovChainState> statiter = states.iterator();
        ArrayList<String> toreturn = new ArrayList<>();
        while (statiter.hasNext()){
            MarkovChainState s = statiter.next();
            VariableList vars = s.getVariables();
            for (MarkovStateVariable var : vars) {
                if (var.isPrivate()==true){
                    MarkovStatePrivateVariable privvar =(MarkovStatePrivateVariable)var;
                    if (privvar.isSecret()){
                        if (toreturn.contains(privvar.getName())==false){
                            toreturn.add(var.getName());
                        }
                    }
                }
            }
        }
        
        return toreturn;
        
    }
        

    protected BigDecimal log2(BigDecimal value) {
      log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision());
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }

    
    public void printGraphToFile(String filename, String type) {
        try {
            File dotFile = new File(filename + ".dot");
            try (PrintWriter out = new PrintWriter(dotFile)) {
                out.println("digraph G {");
                out.println("\trankdir = \"LR\";");
                for (MarkovChainState state : states) {
                    state.printToGraph(out);
                }
                
                out.println("}");
            }
        
            if (type != null) {
                File imgFile = new File(filename + "." + type.substring(2));

                String[] args = {Quail.DOT, type, dotFile.getAbsolutePath(), "-o", imgFile.getAbsolutePath()};
                log("Running: "+args[0]+" "+args[1]+" "+args[2]+" "+args[3]+" "+args[4],1);
                Process p = Runtime.getRuntime().exec(args); 
                p.waitFor();
                dotFile.delete();
            }
        }
        catch (java.io.IOException ioe) {
            System.err.println("Error: in I/O processing files or in calling external command.");
        }
        catch (java.lang.InterruptedException ie) {
            System.err.println("Error: the execution of the external program was interrupted.");
        }
    }

    protected void trimUnreachableStates(){
        log("Trimming unreachable states...",3);
        MarkovStateSet toremove = new MarkovStateSet();
        for (MarkovChainState s : states) {
            if (this.getXiOfState(s).compareTo(contexter.getDecimalZero())==0){
                log("State "+s.getName()+" is unreachable; removing it from the Markov chain",3);
                toremove.add(s);
            }
        }
        if (toremove.isEmpty()==false){
            for (MarkovChainState s : toremove) {
                states.remove(s);
            }
        }
    }
    
    protected final void buildStates(String[] sts, BigDecimal[] initprobs, BigDecimal[][] transs) {
        MarkovStateSet set = new MarkovStateSet();
        for (int i=0; i<sts.length; i++){
            MarkovChainState s = new MarkovChainState(sts[i],initprobs[i]);
            set.add(s);
            
        }
        this.states = set;
        for (int i=0; i<sts.length; i++){
            
            for (int j=0; j<sts.length; j++){
                Transition tr = new Transition(sts[i],sts[j],transs[i][j]);
                this.addTransition(tr);
                if (transs[i][j].compareTo(contexter.getDecimalZero())>0){
                    MarkovChainState source = states.getStateByName(sts[i]);
                    MarkovChainState dest = states.getStateByName(sts[j]);
                    source.addSuccessor(dest);
                    dest.addPredecessor(source);
                }
            }
        }
    }

    protected BigDecimal calculateExpectedReward() {
        BigDecimal expectedreward = contexter.getDecimalZero();
        log("The states are "+states.toString(),4);
        for (MarkovChainState state : states) {
            log("Computing contribution of state "+state.getName()+"...",4);
            BigDecimal contrib = state.getReward();
            if (contrib.compareTo(contexter.getDecimalZero())==0){
                log("State "+state.getName()+" is contributes no reward.",4);
                continue;
            }
            if (this.getXiOfState(state.getName()).compareTo(bogus_infinity) ==0){
                log("ERROR: State "+state.getName()+" is recurrent and has positive reward, it contributes infinite reward.",1);
                continue;
                
            }else{
                contrib = (contrib.multiply(this.getXiOfState(state.getName()),contexter.getContext()));
                if (logger.getVerbosity()>=3){
                    log("State "+state.getName()+" has local reward "+state.getReward().toEngineeringString()+" and expected occupation time "+this.getXiOfState(state.getName()).toEngineeringString()+" and thus contributes a reward of "+contrib.toEngineeringString(),4);
                }
            }
            expectedreward = expectedreward.add(contrib);
        }
        expectedreward = expectedreward.round(contexter.getContext());
        if (logger.getVerbosity()>=3){
            log("The expected reward of the Markov chain is "+expectedreward.toPlainString(),3);
        }
        return expectedreward;
    }
    
    protected BigDecimal calculateExpectedRewardRate() {
        BigDecimal expectedrewardrate = contexter.getDecimalZero();
        log("The states are "+states.toString(),3);
        for (MarkovChainState state : states) {
            log("Computing contribution of state "+state.getName()+"...",3);
            BigDecimal contrib = state.getReward();
            if (contrib.compareTo(contexter.getDecimalZero())==0){
                log("State "+state.getName()+" is contributes no reward.",3);
                continue;
            }
            if (this.getXiOfState(state.getName()).compareTo(bogus_infinity) ==0){
                log("ERROR: State "+state.getName()+" has an infinite rate?",1);
                continue;
                
            }else{
                BigDecimal xirate = this.getXiRateOfState(state.getName());
                contrib = (contrib.multiply(xirate,contexter.getContext()));
                if (logger.getVerbosity()>=3){
                    log("State "+state.getName()+" has local reward "+state.getReward().toEngineeringString()+" and expected occupation time rate "+xirate.toEngineeringString()+" and thus contributes a reward of "+contrib.toEngineeringString(),3);
                }
            }
            expectedrewardrate = expectedrewardrate.add(contrib);
        }
        expectedrewardrate = expectedrewardrate.round(contexter.getContext());
        if (logger.getVerbosity()>=3){
            log("The expected reward rate of the Markov chain is "+expectedrewardrate.toPlainString(),3);
        }
        return expectedrewardrate;
    }

    @Override
    public MarkovChain clone() throws CloneNotSupportedException{
        MarkovChain newchain =(MarkovChain) super.clone();
        newchain.states = this.states.clone();
        return newchain;
    }
    
     
}
