/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.util.HashSet;
import java.util.Iterator;
import java.io.PrintWriter;
import java.math.BigDecimal;
import java.math.BigInteger;

/**
 *
 * @author fbio
 */
public class MarkovChainState {
    private static final Contexter contexter = Contexter.getInstance();
    private String name = "";
    private BigDecimal initialProbability = contexter.getDecimalZero();
    private BigDecimal xi = new BigDecimal(-1.0,contexter.getContext());
    private BigDecimal xirate = new BigDecimal(-1.0,contexter.getContext());
    private BigDecimal localEntropy = new BigDecimal(-1.0,contexter.getContext());
    private BigDecimal reward;
    private BigDecimal probability;
    private boolean recurrent = false;
    private VariableList variables;
    private final MarkovStateSet predecessors;
    private final MarkovStateSet successors;
    private final HashSet<Transition> transitionsFrom;

    

    MarkovChainState(String name, BigDecimal initprob) {
        this.reward = contexter.getDecimalZero();
        this.probability = contexter.getDecimalOne();
        this.name = name;
        this.initialProbability = initprob;
        variables = new VariableList();
        predecessors = new MarkovStateSet();
        successors = new MarkovStateSet();
        transitionsFrom = new HashSet<>();
    }
    
    public BigDecimal getProbability() {
        return probability;
    }

    public void setProbability(BigDecimal probability) {
        this.probability = probability;
    }
    
    public void multiplyProbabilityBy(BigDecimal tomult){
        this.probability = this.probability.multiply(tomult, contexter.getContext());
    }
    
    public void addVariable(MarkovStateVariable var){
        this.variables.add(var);
        
    }
    
    public boolean setPublicVariable(String varname, BigInteger newval){
        return this.variables.setPublicVariable(varname, newval);
    }
    
    public BigInteger getPublicVariableValue(String varname){
        return this.variables.getPublicVariableValue(varname);
    }
    
    public Interval getPrivateVariableInterval(String varname){
        return this.variables.getPrivateVariableInterval(varname);
    }
    
    public boolean deleteVariable(MarkovStateVariable var){
        return this.variables.deleteVariable(var);
    }

    public VariableList getVariables() {
        return variables;
    }

    public void setVariables(VariableList variables) {
        this.variables = variables;
    }
    
    public boolean deleteVariableByName(String name){
        return this.variables.deleteVariableByName(name);
    }
    
    public MarkovStateVariable getVariable(String name){
        return this.variables.getVariable(name);
    }

    public BigDecimal getInitialProbability() {
        return initialProbability;
    }

    public void setInitialProbability(BigDecimal initialProbability) {
        this.initialProbability = initialProbability;
    }

    public BigDecimal getLocalEntropy() {
        return localEntropy;
    }

    public void setLocalEntropy(BigDecimal localEntropy) {
        this.localEntropy = localEntropy;
    }

    public BigDecimal getReward() {
        return reward;
    }
  
    public void setReward(BigDecimal rew){
        this.reward = rew;
    }
    
    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public boolean isRecurrent() {
        return recurrent;
    }

    public void setRecurrent(boolean recurrent) {
        this.recurrent = recurrent;
    }

    public BigDecimal getXi() {
        return xi;
    }

    public MarkovStateSet getPredecessors() {
        return predecessors;
    }

    public void addPredecessor(MarkovChainState predecessor) {
        this.predecessors.add(predecessor);
    }

    public void removePredecessor(MarkovChainState predecessor) {
        this.predecessors.remove(predecessor);
    }

    public MarkovStateSet getSuccessors() {
        return successors;
    }

    public void addSuccessor(MarkovChainState successor) {
        this.successors.add(successor);
    }

    public void removeSuccessor(MarkovChainState successor) {
        this.successors.remove(successor);
    }
    
    public void addTransition(Transition trans){
        this.transitionsFrom.add(trans);
    }

    public void removeTransition(Transition trans){
        this.transitionsFrom.remove(trans);
    }
    
    public void setXi(BigDecimal xi) {
        this.xi = xi;
    }
    
    public BigDecimal getXiRate() {
        return xirate;
    }

    public void setXiRate(BigDecimal xirate) {
        this.xirate = xirate;
    }

    private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
  
    @Override
    public String toString(){
        return this.name;
    }
   
    public String printNice(){
        String s = this.getName();
        if (this.variables.isEmpty()==false){
            s += "{";
            Iterator<MarkovStateVariable> varit = variables.iterator();
            while (varit.hasNext()){
                MarkovStateVariable var =varit.next();
                s += var.toString();
                
                if (varit.hasNext()){
                    s += ",";
                }
            }
            s += "}";
        }
        return s;
    }

    public HashSet<Transition> getTransitionsFrom() {
        return this.transitionsFrom;
    }
    
    public void printToGraph(PrintWriter out) {
        String varstring = "";
        for (MarkovStateVariable var : variables) {
            varstring = varstring + var.toString() + "\\n";
        }

        out.println("\t\"" + this.getName() + "\" [" + "label = \"<f0> " + this.getName() + " | <f1> " + varstring + "\" shape = \"record\"];");
        for (Transition t : transitionsFrom) {
            if (!t.getSource().equals(t.getDest())) {
                out.println("\t\"" + this.getName() + "\":f0 -> \"" + t.getDest() + "\":f0 [label = \"" + t.getValue() + "\"];");
            }
        }
    }

    boolean setPrivateVariable(String varname, String interval) {
        return this.variables.setPrivateVariable(varname, interval);
    }
    
 
    @Override
    public MarkovChainState clone() throws CloneNotSupportedException{
        MarkovChainState copy = new MarkovChainState(this.name,this.initialProbability);
        copy.setVariables(variables.clone());
        copy.setProbability(probability);
        return copy;
    }

    Transition getTransitionToState(MarkovChainState dest) {
        for (Transition trans : this.transitionsFrom) {
            if (trans.getDest().equals(dest.getName())) {
                return trans;
            }
        }
        return new Transition(this.name,dest.getName(),contexter.getDecimalZero());
    }

    void addToReward(BigDecimal toadd) {
        this.reward = this.reward.add(toadd, contexter.getContext());
    }
        
    public VariableList getCloneOfVariables(){
             
        return this.variables.clone();
    }
   
}
