/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012-2014 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/

package quail;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;

/**
 *
 * @author fbio
 */
class PriorDistributionManager {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    
    private ArrayList<String> codelines = null;
    private HashMap<String,Interval> initialsecretintervals = null;

    PriorDistributionManager(ArrayList<String> code) {
        codelines = code;
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
    
    public BigDecimal computePriorShannonEntropy() {
        initialsecretintervals = new HashMap<>();
        BigInteger secretsize = new BigInteger("0");
        for (String line : codelines) {
            if (line.startsWith("secret")){
                if (logger.getVerbosity()>=4){
                    log("Testing if it matches regular expression "+QuailPatterns.pattern_secret.pattern(),4);
                }
                Matcher matcher = QuailPatterns.pattern_secret.matcher(line );
                if ( matcher.find() ) {
                    //log("Num groups: " + matcher.groupCount(),4);
                    log("Variable length: " + matcher.group(1),4);
                    log("Variable name: " + matcher.group(2),4);
                    Integer varlength = Integer.valueOf(matcher.group(1));
                    String varname = matcher.group(2);
                    BigInteger seclength = BigInteger.valueOf(2).pow(varlength.intValue());
                    String varinterval="[0,"+seclength.toString()+"]";
                    initialsecretintervals.put(varname, new Interval(varinterval));
                    log("Secret length: " + seclength,4);
                    if (secretsize.compareTo(BigInteger.ZERO)==0){
                        secretsize = seclength;
                    }else{
                        secretsize = secretsize.multiply(seclength);
                    }
                }else{
                    
                    if (logger.getVerbosity()>=4){
                        log("Testing if it matches regular expression "+QuailPatterns.pattern_secret_with_assignment.pattern(),4);
                    }
                    
                    matcher = QuailPatterns.pattern_secret_with_assignment.matcher(line );
                    if ( matcher.find() ) {
                        //log("Num groups: " + matcher.groupCount(),4);
                        //log("Variable length: " + matcher.group(1),4);
                        log("Variable name: " + matcher.group(2),4);
                        log("Assigned interval: "+matcher.group(3),4);
                        //varlength = Integer.valueOf(matcher.group(1));
                        String varname = matcher.group(2);
                        Interval varinterval = new Interval(matcher.group(3));
   
                        initialsecretintervals.put(varname, varinterval);
                        BigInteger seclength = varinterval.getSize();
                        log("Secret length: " + seclength,4);
                        if (secretsize.compareTo(BigInteger.ZERO)==0){
                            secretsize = seclength;
                        }else{
                            secretsize = secretsize.multiply(seclength);
                        }
                    }
                }
                
            }
 /*           else if (line.startsWith("observable")){
                if (logger.getVerbosity()>=4){
                    log("Testing if it matches regular expression "+pattern_observable.pattern(),4);
                }
                Matcher matcher = pattern_observable.matcher(line );
                if ( matcher.find() ) {
                    //log("Num groups: " + matcher.groupCount(),4);
                    log("Variable length: " + matcher.group(1),4);
                    log("Variable name: " + matcher.group(2),4);
                    Integer varlength = Integer.valueOf(matcher.group(1));
                    //varname = matcher.group(2);
                    BigInteger seclength = BigInteger.valueOf(2).pow(varlength.intValue());
                    //varinterval="[0,"+interend+"]";
                    log("Secret length: " + seclength,4);
                    secretsize.add(seclength);
                }else{
                    
                    if (logger.getVerbosity()>=4){
                        log("Testing if it matches regular expression "+pattern_secret_with_assignment.pattern(),4);
                    }
                    
                    matcher = pattern_secret_with_assignment.matcher(line );
                    if ( matcher.find() ) {
                        //log("Num groups: " + matcher.groupCount(),4);
                        //log("Variable length: " + matcher.group(1),4);
                        log("Variable name: " + matcher.group(2),4);
                        log("Assigned interval: "+matcher.group(3),4);
                        //varlength = Integer.valueOf(matcher.group(1));
                        //  varname = matcher.group(2);
                        Interval varinterval = new Interval(matcher.group(3));
                        BigInteger seclength = varinterval.getSize();
                        log("Secret length: " + seclength,4);
                        secretsize.add(seclength);
                    }
                }
            }*/
            
        }
        BigDecimal decsecretsize = new BigDecimal(secretsize);
        BigDecimal priorentropy = log2(decsecretsize);
        log("Secret size as decimal: value "+decsecretsize.toPlainString()+", scale: "+decsecretsize.scale()+", base 2 logarithm "+priorentropy.toPlainString(),4);
         
         log("Total secret size: "+secretsize.toString()+" different values, corresponding to "+priorentropy+" bits.",2);
         return priorentropy;
    }

    HashMap<String, Interval> getInitialSecretIntervals() {
        if (initialsecretintervals == null){
            this.computePriorShannonEntropy();
        }
        return this.initialsecretintervals;
    }
    
    private BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision());
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
        
      
    }
    
}
