/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012-2014 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.io.File;
import java.io.FileNotFoundException;
import static java.lang.Math.max;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.Scanner;
import preprocessor.Preprocessor;


/**
 *
 * @author fbio
 */
public class Quail {
    
    protected static String DOT = "/usr/bin/dot";
    

    
    private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
    
    
    private static void printHelp() {
        log("Quail 2.0 (Quantitative Analyzer for Imperative Languages)",0);
        log("",0);
        log("Description:",0);
        log("    QUAIL analyzes a program written in an imperative language with a quantitative security analysis",0);
        log("    that builds andevaluates observability models, and finally computes the information leakage",0);
        log("    of the program.",0);
        log("",0);
        log("Usage:",0);
        log("    quail [options] program.quail",0);
        log("",0);
        log("Options:",0);
        log("    -c | --clean                 Delete the temporary preprocessed file.",0);
        log("    --config=filename            Alternative location for the configuration file.",0);
        log("    -h | -? | --help             Print this help and terminate.",0);
        log("    -v level                     Verbose mode between 0 (minimum) and 4 (maximum).",0);
        log("                                 Default value if not specified is 2.",0);
        log("    -th number                   Sets the number of parallel threads QUAIL should use. Default is the number",0);
        log("                                 returned by Runtime.getRuntime().availableProcessors(), minimum 1",0);
        log("    -p value                     Define computation precision as a number of digits.",0);
        log("                                 Default value if not specified is 15.",0);
        log("    --const:<cname>:=<cvalue>    Replace the value of a constant cname in the model by the value cvalue",0);
        log("",0);
        log("    --preprocess                 Run only the preprocessor and terminate.",0);
        log("",0);
        log("    --simulate                   Run the simulator instead of the analyzer.",0);
        log("                                 If this option is chosen, the following additional options are available:",0);
        log("                                 -i iterations       Sets the number of simulations. Default: 50000",0);
        log("                                 -l threshold        Sets the sufficient probability threshold after which",0);
        log("                                                     exploration of the probability space can be terminated.",0);   
        log("                                                     Default: 1 (meaning 100% of the probability space)",0);
        log("                                 -t time_limit       Sets the time limit (in milliseconds) for the simulation.",0);
        log("                                                     A value of 0 means no time limit.",0);   
        log("                                                     Default: 0 (meaning no time limit)",0);
        log("",0);
        log("    --fullmc                     Runs the QUAIl version 1 algorithm constructing the full Markov chain (slow).",0);
        log("                                 If this option is chosen, the following additional options are available:",0);
        log("                                 --no_hiding          Do not run the hiding algorithm.",0);
        log("                                 -mc [values]  Print the Markov chains:" , 0);
        log("                                               - 0 program Markov chain",0);
        log("                                               - 1 hiding non observable states",0);
        log("                                               - 2 observer's quotient",0);
        log("                                               - 3 secret's quotient",0);      
        log("                                               - 4 joint quotient",0);
        log("                                               e.g. 0 1 2 3 4 will print all Markov chains.",0);
        log("                                               By default, if no value is given, print the program Markov chain.",0);
        log("                                               The output format can be any format supported by Graphviz,",0);
        log("                                               Use option -Tgif,-Tfig,-Tpdf,-Tps,-Tsvg or -Tpng for the most common.",0);
        log("                                               By default if no format is provided it prints in dot format.",0);
        log("                                               -o filename   File name for Markov chains. Suffixes are added automatically." , 0);
        log("",0);
        

        
    }

    public static void main(String args[]) {
        
        // verbose mode between 0 and 4
        int verbose = 2;
        
        // precision
        int precision = 15;
        
        // number of threads
        int threads = max(1,Runtime.getRuntime().availableProcessors());
        
        // iterations for the simulator
        int iterations = 50000;
        
        // treshold for the simulator to terminate
        BigDecimal threshold = new BigDecimal("1");
        
        // time limit for the simulator
        long timelimit = 0;
        // Markov chains to print
        String [] printMc = {null,null,null,null,null,null};
        
        // tasks to perform
        boolean run_preprocessor = true;
        boolean run_analyzer = false;
        boolean run_hiding = true;
        boolean run_simulator = false;
        boolean run_rate = false;
        boolean run_explorer = true;
        boolean add_free = false;
        boolean clean = false;
  
        
        // files
        String config ="quail.config";
        String filename = "";
        String outfilename = "";
        
        // constants
        HashMap<String,BigDecimal> constants = new HashMap<>();
        
        // parse command line options
        String arg;
        for( int i=0; i<args.length; i++) {
            arg = args[i];
            
            // help
            if (arg.equals("-h")) {
                printHelp();
                System.exit(1);
            }
            else if (arg.equals("-?")) {
                printHelp();
                System.exit(1);
            }
            else if (arg.equals("--help")) {
                printHelp();
                System.exit(1);
            }
            // verbose
            else if (arg.equals("-v")) {
                i++;
                verbose = Integer.parseInt(args[i]);
                if ((verbose<0)||(verbose>4)){
                    log("ERROR: verbosity must be between 0 and 4",1);
                    System.exit(1);
                }
            }
            // number of threads
            else if (arg.equals("-th")) {
                i++;
                threads = Integer.parseInt(args[i]);
                if ((threads<1)){
                    log("ERROR: number of threads must be at least 1",1);
                    System.exit(1);
                }
            }
            
            // iterations
            else if (arg.equals("-i")) {
                i++;
                iterations = Integer.parseInt(args[i]);
                if (iterations<1){
                    log("ERROR: at least 1 iteration needed",1);
                    System.exit(1);
                }
            }
            // precision
            else if (arg.equals("-p")) {
                i++;
                precision = Integer.parseInt(args[i]);
                if (precision<1){
                    log("ERROR: at least 1 digit precision needed",1);
                    System.exit(1);
                }
                if (precision<5){
                    log("WARNING: at least 5 digit precision recommended",2);
                    
                }
            }
            // probability threshold for the simulation
            else if (arg.equals("-l")) {
                i++;
                try{
                threshold = new BigDecimal(args[i]);
                } catch (NumberFormatException e){
                    log("ERROR: unable to convert '"+args[i]+"' in a BigDecimal threshold",1);
                    System.exit(1);
                }
                if (threshold.compareTo(new BigDecimal(0))<=0){
                    log("ERROR: threshold must be larger than zero",1);
                    System.exit(1);
                }
                if (threshold.compareTo(new BigDecimal(1))>0){
                    log("ERROR: threshold must be smaller or equal than one",1);
                    System.exit(1);
                }
            }
            // time limit for the simulation
           else if (arg.equals("-t")) {
                i++;
                timelimit = Long.parseLong(args[i]);
                if (timelimit<0){
                    log("ERROR: time limit should be a positive number of milliseconds, or 0 for no limit",1);
                    System.exit(1);
                }
            }
            // print markov chains
            else if (arg.equals("-mc")) {
                boolean hasValues = false; 
                boolean stillHasValues = true;
                do { 
                    i++;
                    try {
                        int val = Integer.parseInt(args[i]);
                        if (val >= 0 && val <= 4) {
                            printMc[val+1] = "";
                            hasValues = true;
                        }
                        else {
                            System.err.println("Uncompatible values for option -mc: "+val);
                            printHelp();
                            System.exit(1); 
                        }
                    }
                    catch (NumberFormatException e) {
                        stillHasValues = false;
                        i--;
                    }
                } while (stillHasValues);
                
                if(!hasValues) {
                    printMc[1] = "";
                }
            }
            // print format
            else if (arg.startsWith("-T")) {
                printMc[0]=arg;
            }
            else if (arg.equals("-o")) {
                i++;
                outfilename = args[i];
            }


            // preprocessor only
            else if (arg.equals("--preprocess")) {
                run_analyzer = false;
                run_simulator = false;
                run_explorer = false;
            }
            // old algorithm
            else if (arg.equals("--fullmc")) {
                run_analyzer = true;
                run_simulator = false;
                run_explorer = false;
                add_free = true;
            }
                 
            // simulate only
            else if (arg.equals("--simulate")) {
                run_explorer = false;
                run_simulator = true;
                
            }
            /*
            
            else if (arg.equals("--rate")) {
                run_rate = true;
            }
            */
            else if (arg.equals("--no_hiding")) {
                run_hiding = false;
            }
            
            // constants            
            else if (arg.startsWith("--const:")) {
                String[] constant = arg.substring(8).split(":=");
                constants.put(constant[0],new BigDecimal(constant[1]));
            }
            
            // clean
            else if (arg.equals("-c") || arg.equals("--clean")) {
                clean = true;
            }
            // configuration file
            else if (arg.substring(0,9).equals("--config=")) {
                config = arg.substring(9);
            }
            else {
                filename = arg;
            }
        }
        
        // read configuration file
        try {
            Scanner scanner = new Scanner(new File(config));
            while (scanner.hasNextLine()){
                String line = scanner.nextLine();
                if (line.startsWith("DOT=")) {
                    Quail.DOT = line.substring(4,line.length());
                }
            }
        }
        catch (FileNotFoundException e) {
            System.err.println("Error: invalid configuration filename: " + config);
            System.err.println("Default values are assumed.");
        }
        
        // read file
        File f = new File(filename);
        if (!f.exists()) {
            System.err.println("Error: invalid filename " + filename);
            System.exit(1);
        }
        
        // configure output files
        if (outfilename.isEmpty()) {
            outfilename = filename;
        }
        for (int i = 1; i < printMc.length; i++) {
            if (printMc[i] != null) {
                printMc[i] = outfilename + ".mc" + (i-1);
            }
        }
              
        Logger logger = Logger.getInstance();
        logger.setVerbosity(verbose);
        log("Setting verbosity to "+verbose,4);
        
        MathContext context = new MathContext(precision,RoundingMode.HALF_EVEN);
        Contexter contexter = Contexter.getInstance();
        contexter.setContext(context);
        log("Setting precision to "+precision,4);
        //BigDecimal deczero = contexter.getDecimalZero();
        //log("Decimal zero is "+deczero.toPlainString()+" with precision "+deczero.precision()+" and scale "+deczero.scale(),4);

        // preprocessing
        if (run_preprocessor) {
	    
            try {
                Preprocessor.preprocess(filename,constants,add_free);
            }
            catch(Exception e) {
                System.err.println("Error while preprocessing program. Aborting. ");
                e.printStackTrace(System.err);
                System.exit(1);
            }
            filename += ".pp";
            f = new File(filename);
        }

        // quail

       /* if (run_analyzer) {
            Analyzer anal = new Analyzer();
            anal.analyze(filename,printMc);
        }*/
        if (run_rate){
            run_analyzer = true;
            run_simulator = false;
            run_hiding = false;
        }
        
        if (run_simulator){
            run_analyzer = false;
        }

        if (run_analyzer) {
            Analyzer.analyze(filename,printMc,run_hiding,run_rate);
        }
        
        if (run_simulator){
            ConcurrentSimulator sim = new ConcurrentSimulator(iterations,threshold,timelimit,threads);
            sim.simulate(filename);
        }
        
        if (run_explorer){
                ConcurrentExplorer exp = new ConcurrentExplorer(threads);
                exp.explore(filename);
            
            
        }
        
        // clean
        if (run_preprocessor && clean) {
            f.delete();
        }
    }   
}
