/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012-2014 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/

package quail;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.Set;

/**
 *
 * @author fbio
 */
class SimpleMultiIntervalProbabilityDistribution extends HashMap< IntervalAssignments , BigDecimal >{
    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
  
    public String printNice() {
        String toreturn = "";
        Set<IntervalAssignments> keys = this.keySet();
        for (IntervalAssignments key : keys){
            
            toreturn +="assignment "+key.printNice()+" with probability "+this.get(key).toEngineeringString()+"\n";
        }
        return toreturn;
    }

    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }

    void update(VariableList secrets, BigDecimal stateprob) {
        this.update(new IntervalAssignments(secrets),stateprob);
    }
    
    synchronized void update(IntervalAssignments secrets, BigDecimal stateprob) {
        IntervalAssignments newassignment = secrets;
        BigDecimal prob = stateprob;
        
        log("Updating with assignment "+secrets.printNice()+" and probability "+stateprob.toEngineeringString(),3);
        Set<IntervalAssignments> keys = this.keySet();
        // if the new assignment already exists, just update its value
        for (IntervalAssignments oldassignment : keys){
            if (oldassignment.equals(newassignment)){
                log("The assignment already exists; updating its probability",3);
                this.addProbability(oldassignment,prob);
                return;
            }
        }
        
        HashMap< IntervalAssignments , BigDecimal > newdist = new HashMap();
        
        
        for (IntervalAssignments oldassignment : keys){
            if (newassignment.hasEmptyInterval()){
                newdist.put(oldassignment, this.get(oldassignment));
                continue;
            }
            //log("Old HFP: "+cloneoldHFP.printNice(),3);
            // we have the old assignment and the new assignment; we obtain their intersection, the rest of the old assignment without the intersection, and the rest of the new assignment without the interseciton
            IntervalAssignments[] splitted = oldassignment.splitIntervals(newassignment);
            log("Splitting current assignment "+oldassignment.printNice()+" with new assignment "+newassignment.printNice(),3);
            IntervalAssignments intersection = splitted[0];
            IntervalAssignments restofold    = splitted[1];
            IntervalAssignments restofnew    = splitted[2];
            log("Result of split:\nintersection: "+intersection.printNice()+"\nrest of old: "+restofold.printNice()+"\nrestofnew: "+restofnew.printNice(),3);
        
            BigInteger oldsize = oldassignment.getSize();
            BigInteger intsize = intersection.getSize();
            BigInteger newsize = newassignment.getSize();
            
            // old interval: change probability to account for reduced size and put with same frequency
            if (!restofold.hasEmptyInterval()){
                BigDecimal newprob = this.get(oldassignment);
                BigInteger restofoldsize = restofold.getSize();
                if (oldsize!=restofoldsize){
                    BigDecimal tomult = new BigDecimal(restofoldsize,contexter.getContext());
                    tomult = tomult.divide(new BigDecimal(oldsize,contexter.getContext()), contexter.getContext());
                    newprob = newprob.multiply(tomult, contexter.getContext());
                    
                }
                log("The old interval minus the intersection is "+restofold.printNice()+", putting it back in with probability "+newprob,3);
                newdist.put(restofold, newprob);
            }else{
                log("The old interval minus the intersection has an empty interval",3);
            }
            
            // intersection: take probability by size from old interval, then update with prob by size from new interval (freq is freq of old plus one)
            if (!intersection.hasEmptyInterval()){
                
                BigDecimal intersprob =  this.get(oldassignment);
                if (oldsize!=intsize){
                    BigDecimal tomult = new BigDecimal(intsize).divide(new BigDecimal(oldsize), contexter.getContext());
                    intersprob = intersprob.multiply(tomult,contexter.getContext());
                }
                BigDecimal toadd = prob;
                toadd = toadd.multiply(new BigDecimal(intsize,contexter.getContext()), contexter.getContext());
                toadd = toadd.divide(new BigDecimal(newsize,contexter.getContext()), contexter.getContext());
                
                intersprob = intersprob.add(toadd, contexter.getContext());
                
                log("The intersection is "+intersection.printNice()+", putting it in in with probability "+intersprob.toEngineeringString(),3);
                newdist.put(intersection, intersprob);
                
                // change remaining probability to reflect the fact that some of it has gone to the intersection
                prob = prob.subtract(toadd, contexter.getContext());
                
            }else{
                log("The intersection has an empty interval",3);
            }
            
            
            // new interval: change probability to account for reduced size, wait for next iteration
            newassignment = restofnew;
            log("The remainder is "+newassignment.printNice()+" with probability "+prob.toEngineeringString(),3);
            
        }
        // if new interval is now not empty, put it with the remaining probability and frequency 1
        
        
        if (!newassignment.hasEmptyInterval()){
            log("Inserting the remainder "+newassignment.toString()+" with probability "+prob.toPlainString(),3);
            
            newdist.put(newassignment, prob);
        }
        this.clear();
        this.putAll(newdist);
    }

    void initialize(VariableList secrets, BigDecimal stateprob) {
        IntervalAssignments newkey = new IntervalAssignments(secrets);
        
       
        
        this.put(newkey, stateprob);
        
    }


    

    public BigDecimal getEntropy() {
        BigDecimal entropy = contexter.getDecimalZero();
        for (IntervalAssignments inter : this.keySet()){
            BigInteger intervsize = inter.getSize();
            
            
            BigDecimal prob = this.get(inter);
            BigDecimal localent = prob.multiply(log2(prob.divide(new BigDecimal(intervsize), contexter.getContext())),contexter.getContext());
            
            log("Local entropy: "+localent,4);
            entropy = entropy.subtract(localent, contexter.getContext());

        }
        return entropy;

    }
    
    protected BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision()+10);
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }

    public void normalizeTo(BigDecimal prob) {
        BigDecimal currprob = this.getTotalProbability();
        if (currprob.compareTo(prob)!=0){
            BigDecimal multiplier = prob.divide(currprob, contexter.getContext());
            for (IntervalAssignments interv : this.keySet()){
                this.put(interv, this.get(interv).multiply(multiplier, contexter.getContext()));
            }
        }
    }
    
    public void normalize() {
        this.normalizeTo(contexter.getDecimalOne());

    }

    private BigDecimal getTotalProbability() {
        BigDecimal toreturn = contexter.getDecimalZero();
        for (BigDecimal prob: this.values()){
            toreturn = toreturn.add(prob, contexter.getContext());
        }
        return toreturn;
    }

    private void addProbability(IntervalAssignments newassignment, BigDecimal prob) {
        BigDecimal currprob = this.get(newassignment);
        this.put(newassignment, currprob.add(prob, contexter.getContext()));
    }
}
