/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

/**
 *
 * @author fbio
 */
public class Tarjan {
    
  private int index = 0;
  private final ArrayList<Node> stack = new ArrayList<>();
  private final HashSet<HashSet<String>> SCC = new HashSet<>();
  private final NodeList nodes = new NodeList();
  private final EdgeList edges = new EdgeList();
  /*
    public Tarjan(MarkovChain mc) {
    }
    
    public Tarjan(IntervalMarkovChain imc) {
    }
*/
  /* The function tarjan has to be called for every unvisited node of the graph */
  public NodeList getNodes(){
      return this.nodes;
  }
  
  public EdgeList getEdges(){
      return this.edges;
  }
  
  public boolean initialize(MarkovChain mc){
      log("Initializing Tarjan for a Markov chain with "+mc.getStates().size()+" states",4);
      HashSet<Transition> transitions = mc.getTransitions();
      Iterator<Transition> transiter = transitions.iterator();
      HashSet<String> knownnodes = new HashSet<>();
       while (transiter.hasNext()){
           Transition trans =transiter.next();
           Node nodefrom = new Node (trans.getSource());
           Node nodeto = new Node (trans.getDest());
           Edge newedge = new Edge(nodefrom.getElement(),nodeto.getElement());
           if (trans.getValue().compareTo(BigDecimal.ZERO)>0){
               edges.add(newedge);
           }
           if (knownnodes.contains(nodefrom.getElement())==false) {
               nodes.add(nodefrom);
               knownnodes.add(nodefrom.getElement());
           }
           if (knownnodes.contains(nodeto.getElement())==false) {
               nodes.add(nodeto);
               knownnodes.add(nodeto.getElement());
           }
       }
      log("The nodes are "+nodes.toString(),4);
      log("The edges are "+edges.toString(),4);
      if (nodes.size()!=mc.getStates().size()){
          log("ERROR: Tarjan not initialized correctly: Markov chain has "+mc.getStates().size()+" nodes but "+nodes.size()+" were added",1);
          return false;
      }//more things to check?
      log("Tarjan initialized correctly",4);
      return true;  
  }
  
/*  public boolean initialize(IntervalMarkovChain imc){
 //     log("Initializing Tarjan for an Interval Markov chain with "+imc.getStates().size()+" states and "+imc.getTransitions().size()+" transitions");
      HashSet transitions = imc.getTransitions();
      Iterator transiter = transitions.iterator();
      HashSet knownnodes = new HashSet();
       while (transiter.hasNext()){
           IntervalTransition trans =(IntervalTransition) transiter.next();
           Node nodefrom = new Node (trans.getSource());
           Node nodeto = new Node (trans.getDest());
           Edge newedge = new Edge(nodefrom.getElement(),nodeto.getElement());
           if (trans.getTop()>0.0) edges.add(newedge);
           if (knownnodes.contains(nodefrom.getElement())==false) {
               nodes.add(nodefrom);
               knownnodes.add(nodefrom.getElement());
           }
           if (knownnodes.contains(nodeto.getElement())==false) {
               nodes.add(nodeto);
               knownnodes.add(nodeto.getElement());
           }
       }
      //log("The nodes are "+nodes.toString());
      //log("The edges are "+edges.toString());
      if (nodes.size()!=imc.getStates().size()){
          log("ERROR: Tarjan not initialized correctly: Markov chain has "+imc.getStates().size()+" nodes but "+nodes.size()+" were added");
          return false;
      }//more things to check?
  //    log("Tarjan initialized correctly");
      return true;  
      
  }
  */
  public HashSet<HashSet<String>> getSCC(){
       SCC.clear();
       index = 0;
       stack.clear();
       if(edges != null){
          ArrayList<Node> nodeList = new ArrayList<>();
          nodeList.addAll(nodes);
          if(nodeList != null)
          {
              for (Node node : nodeList) {
                  log("Selected node "+node.printAll(),4);
                  if(node.index == -1)
                  {
                      log("Node "+node+" does not have an index; calling Tarjan's algorithm for node "+node.printAll()+" with edges "+edges.getEdgesFrom(node),4);
                      tarjan(node);
                      log("Now it's node "+node.printAll(),4);
                  }else {
                      log("Already visited; moving on",4);
                  }
              }
           }
       }
     return SCC;
  }
   private HashSet<HashSet<String>> tarjan(Node v){
       v.index = index;
       v.lowlink = index;
       log("I gave the index "+index+" to node "+v,4);
       index++;
       stack.add(0, v);
       EdgeList edgesfromv = edges.getEdgesFrom(v);
       log("The edges from "+v+" are "+edgesfromv,4);
      for (Edge e : edgesfromv) {
          Node n = nodes.getNodeByName(e.to);
          if(n.index == -1){
              log("Node "+n.printAll()+" successor of "+v.printAll()+" has no index, calling Tarjan's algorithm recursively for it",4);
              //log("Here list has "+list.size()+" elements",4);
              tarjan(n);
              v.lowlink = Math.min(v.lowlink, n.lowlink);
              
          }else if(stack.contains(n)){
              v.lowlink = Math.min(v.lowlink, n.index);
          }
          log("Now node "+v+" is "+v.printAll(),4);
      }
       if(v.lowlink == v.index){
           log("Node "+v+" is the root of a SCC!",4);
           Node n;
           HashSet<String> component = new HashSet<>();
           do{
               n =stack.remove(0);
               component.add(n.getElement());
               //log("Adding "+n+" to the SCC of "+v);
           }while(n != v);
           SCC.add(component);
       }
       return SCC;
   }

 
   private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
}
