/*  QUAIL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2012 - INRIA Rennes (FR) and IT University of Copenhagen (DK)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Fabrizio Biondi <fabrizio.biondi@inria.fr>
*/
    
package quail;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;


/**
 *
 * @author fbio
 */
class VariableList extends ArrayList<MarkovStateVariable>{

    public VariableList() {
    }
    
    
 
    
    public MarkovStateVariable getVariable(String name){
        MarkovStateVariable toreturn = null;
        for (MarkovStateVariable temp : this) {
            if (temp.getName().equals(name)){
                toreturn = temp;
                break;
            }
        }
        //if (toreturn==null){
        //    log("WARNING: asked to get non-existent variable "+name+" from state "+this.getName()+", returning null instead");
        //}
        return toreturn;
    } 
    
    public boolean deleteVariableByName(String name){
        MarkovStateVariable var = this.getVariable(name);
        if (var==null) {            
            log("WARNING: attempted to delete non-existent variable "+name,1);
            return false;
        }else{
            
            return this.deleteVariable(var);
        }
            
    }
    public boolean deleteVariable(MarkovStateVariable var){
        if (this.contains(var)){
            this.remove(var);
            return true;
        }else{
            log("WARNING: attempted to delete non-existent variable "+var.getName(),1);
            return false;
        }
    }
    
    public boolean setPublicVariable(String varname, BigInteger newval) {
        MarkovStateVariable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to assign value "+newval+" to variable "+varname+", but the variable does not exist",1);
            return false;
        }else if (var.isPrivate()){
            log("ERROR: attempt to assign value "+newval+" to variable "+varname+", but the variable is private",1);
            return false;
        }else{
            MarkovStatePublicVariable pubvar =(MarkovStatePublicVariable) var;
            pubvar.setVal(newval);
            return true;
        }
    }
    
    public boolean setPublicVariable(String varname, long newval){
       return this.setPublicVariable(varname, BigInteger.valueOf(newval));
    }
    
    public BigInteger getPublicVariableValue(String varname){
        MarkovStateVariable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to get value of public variable "+varname+", but the variable does not exist",1);
            return BigInteger.valueOf(-1);
        }else if (var.isPrivate()){
            log("ERROR: attempt to get value of public variable "+varname+", but the variable is private",1);
            return BigInteger.valueOf(-1);
        }else{
            MarkovStatePublicVariable pubvar =(MarkovStatePublicVariable) var;
            return pubvar.getVal();
        }
    }
    
    public Interval getPrivateVariableInterval(String varname){
        MarkovStateVariable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to get interval of private variable "+varname+", but the variable does not exist",1);
            return null;
        }else if (var.isPrivate()==false){
            log("ERROR: attempt to get interval of private variable "+varname+", but the variable is public",1);
            return null;
        }else{
            MarkovStatePrivateVariable privvar =(MarkovStatePrivateVariable) var;
            return privvar.getInterval();
        }
    }
    
    public String getIdentifier(){
        this.sort();
        return this.toString();
    }    
    
    
    public VariableList clone(){
        VariableList toreturn =new VariableList();
        for (MarkovStateVariable myvar : this) {
            if (myvar.isPrivate()){
                MarkovStatePrivateVariable myprivvar = (MarkovStatePrivateVariable)myvar;
                MarkovStatePrivateVariable newprivvar = new MarkovStatePrivateVariable(myprivvar.getName(),myprivvar.getInterval().toString(),myprivvar.getLength());
                newprivvar.setSecret(myprivvar.isSecret());
                toreturn.add(newprivvar);
            }else{
                MarkovStatePublicVariable mypubvar = (MarkovStatePublicVariable)myvar;
                MarkovStatePublicVariable newpubvar = new MarkovStatePublicVariable(mypubvar.getName(),mypubvar.getVal(),mypubvar.getLength());
                newpubvar.setObservable(mypubvar.isObservable());
                toreturn.add(newpubvar);
            }
        }
        return toreturn;
        
    }
   
    
    @Override
    public String toString(){
        String s="";
        for (MarkovStateVariable var : this) {
            s+=var.toString()+";";
        }
        return s;
    }
    
    private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }

    

    public boolean setPrivateVariable(String varname, String interval) {
        MarkovStateVariable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to assign interval "+interval+" to variable "+varname+", but the variable does not exist",1);
            return false;
        }else if (!var.isPrivate()){
            log("ERROR: attempt to assign interval "+interval+" to variable "+varname+", but the variable is public",1);
            return false;
        }else{
            MarkovStatePrivateVariable privvar =(MarkovStatePrivateVariable) var;
            privvar.setInterval(interval);
            return true;
        }
    }

    private void sort() {
        Collections.sort(this, new MarkovStateVariableComparator());
    }

    
    }

    
    
    

